# SPDX-License-Identifier: Apache-2.0
# Copyright © 2020-2025 Intel Corporation

from __future__ import annotations
import itertools
import os
import re
import textwrap
import typing as T

from mesonbuild.interpreterbase.decorators import FeatureNew

from . import ExtensionModule, ModuleReturnValue, ModuleInfo
from .. import mesonlib, mlog
from ..build import (BothLibraries, BuildTarget, CustomTargetIndex, Executable, ExtractedObjects, GeneratedList,
                     CustomTarget, InvalidArguments, Jar, StructuredSources, SharedLibrary, StaticLibrary)
from ..compilers.compilers import are_asserts_disabled_for_subproject, lang_suffixes
from ..interpreter.type_checking import (
    DEPENDENCIES_KW, LINK_WITH_KW, LINK_WHOLE_KW, SHARED_LIB_KWS, TEST_KWS, TEST_KWS_NO_ARGS,
    OUTPUT_KW, INCLUDE_DIRECTORIES, SOURCES_VARARGS, NoneType, in_set_validator
)
from ..interpreterbase import ContainerTypeInfo, InterpreterException, KwargInfo, typed_kwargs, typed_pos_args, noPosargs, permittedKwargs
from ..interpreter.interpreterobjects import Doctest
from ..mesonlib import File, MesonException, PerMachine
from ..programs import ExternalProgram, NonExistingExternalProgram

if T.TYPE_CHECKING:
    from . import ModuleState
    from ..build import IncludeDirs, LibTypes
    from ..compilers.rust import RustCompiler
    from ..dependencies import Dependency, ExternalLibrary
    from ..interpreter import Interpreter
    from ..interpreter import kwargs as _kwargs
    from ..interpreter.interpreter import SourceInputs, SourceOutputs
    from ..interpreter.interpreterobjects import Test
    from ..programs import OverrideProgram
    from ..interpreter.type_checking import SourcesVarargsType

    from typing_extensions import TypedDict, Literal

    ArgsType = T.TypeVar('ArgsType')

    class FuncRustTest(_kwargs.BaseTest, T.Generic[ArgsType]):
        args: T.List[ArgsType]
        dependencies: T.List[T.Union[Dependency, ExternalLibrary]]
        is_parallel: bool
        link_with: T.List[LibTypes]
        link_whole: T.List[LibTypes]
        rust_args: T.List[str]

    FuncTest = FuncRustTest[_kwargs.TestArgs]
    FuncDoctest = FuncRustTest[str]

    class FuncBindgen(TypedDict):

        args: T.List[str]
        c_args: T.List[str]
        include_directories: T.List[IncludeDirs]
        input: T.List[SourceInputs]
        output: str
        output_inline_wrapper: str
        dependencies: T.List[T.Union[Dependency, ExternalLibrary]]
        language: T.Optional[Literal['c', 'cpp']]
        bindgen_version: T.List[str]


RUST_TEST_KWS: T.List[KwargInfo] = [
     KwargInfo(
         'rust_args',
         ContainerTypeInfo(list, str),
         listify=True,
         default=[],
         since='1.2.0',
     ),
     KwargInfo('is_parallel', bool, default=False),
]

def no_spaces_validator(arg: T.Optional[T.Union[str, T.List]]) -> T.Optional[str]:
    if any(bool(re.search(r'\s', x)) for x in arg):
        return 'must not contain spaces due to limitations of rustdoc'
    return None


class RustModule(ExtensionModule):

    """A module that holds helper functions for rust."""

    INFO = ModuleInfo('rust', '0.57.0', stabilized='1.0.0')
    _bindgen_rust_target: T.Optional[str]
    rustdoc: PerMachine[T.Optional[ExternalProgram]] = PerMachine(None, None)

    def __init__(self, interpreter: Interpreter) -> None:
        super().__init__(interpreter)
        self._bindgen_bin: T.Optional[T.Union[ExternalProgram, Executable, OverrideProgram]] = None
        if 'rust' in interpreter.compilers.host:
            rustc = T.cast('RustCompiler', interpreter.compilers.host['rust'])
            self._bindgen_rust_target = 'nightly' if rustc.is_nightly else rustc.version
        else:
            self._bindgen_rust_target = None
        self._bindgen_set_std = False
        self.methods.update({
            'test': self.test,
            'doctest': self.doctest,
            'bindgen': self.bindgen,
            'proc_macro': self.proc_macro,
        })

    def test_common(self, funcname: str, state: ModuleState, args: T.Tuple[str, BuildTarget], kwargs: FuncRustTest) -> T.Tuple[Executable, _kwargs.FuncTest]:
        """Generate a rust test target from a given rust target.

        Rust puts its unitests inside its main source files, unlike most
        languages that put them in external files. This means that normally
        you have to define two separate targets with basically the same
        arguments to get tests:

        ```meson
        rust_lib_sources = [...]
        rust_lib = static_library(
            'rust_lib',
            rust_lib_sources,
        )

        rust_lib_test = executable(
            'rust_lib_test',
            rust_lib_sources,
            rust_args : ['--test'],
        )

        test(
            'rust_lib_test',
            rust_lib_test,
            protocol : 'rust',
        )
        ```

        This is all fine, but not very DRY. This method makes it much easier
        to define rust tests:

        ```meson
        rust = import('unstable-rust')

        rust_lib = static_library(
            'rust_lib',
            [sources],
        )

        rust.test('rust_lib_test', rust_lib)
        ```
        """
        if any(isinstance(t, Jar) for t in kwargs.get('link_with', [])):
            raise InvalidArguments('Rust tests cannot link with Jar targets')
        if any(isinstance(t, Jar) for t in kwargs.get('link_whole', [])):
            raise InvalidArguments('Rust tests cannot link with Jar targets')

        name = args[0]
        base_target: BuildTarget = args[1]
        if not base_target.uses_rust():
            raise InterpreterException(f'Second positional argument to rustmod.{funcname}() must be a rust based target')
        extra_args = kwargs['args']

        # Delete any arguments we don't want passed
        if '--test' in extra_args:
            mlog.warning(f'Do not add --test to rustmod.{funcname}() arguments')
            extra_args.remove('--test')
        if '--format' in extra_args:
            mlog.warning(f'Do not add --format to rustmod.{funcname}() arguments')
            i = extra_args.index('--format')
            # Also delete the argument to --format
            del extra_args[i + 1]
            del extra_args[i]
        for i, a in enumerate(extra_args):
            if isinstance(a, str) and a.startswith('--format='):
                del extra_args[i]
                break

        # We need to cast here, as currently these don't have protocol in them, but test itself does.
        tkwargs = T.cast('_kwargs.FuncTest', kwargs.copy())

        tkwargs['args'] = extra_args + ['--test', '--format', 'pretty']
        tkwargs['protocol'] = 'rust'

        new_target_kwargs = base_target.original_kwargs.copy()
        # Don't mutate the shallow copied list, instead replace it with a new
        # one
        new_target_kwargs['install'] = False
        new_target_kwargs['dependencies'] = new_target_kwargs.get('dependencies', []) + kwargs['dependencies']
        new_target_kwargs['link_with'] = new_target_kwargs.get('link_with', []) + kwargs['link_with']
        new_target_kwargs['link_whole'] = new_target_kwargs.get('link_whole', []) + kwargs['link_whole']
        del new_target_kwargs['rust_crate_type']
        for kw in ['pic', 'prelink', 'rust_abi', 'version', 'soversion', 'darwin_versions']:
            if kw in new_target_kwargs:
                del new_target_kwargs[kw]

        lang_args = base_target.extra_args.copy()
        lang_args['rust'] = base_target.extra_args['rust'] + kwargs['rust_args'] + ['--test']
        new_target_kwargs['language_args'] = lang_args

        sources = T.cast('T.List[SourceOutputs]', base_target.sources.copy())
        sources.extend(base_target.generated)

        new_target = Executable(
            name, base_target.subdir, state.subproject, base_target.for_machine,
            sources, base_target.structured_sources,
            base_target.objects, base_target.environment, base_target.compilers,
            new_target_kwargs
        )
        return new_target, tkwargs

    @typed_pos_args('rust.test', str, BuildTarget)
    @typed_kwargs(
        'rust.test',
        *TEST_KWS,
        DEPENDENCIES_KW,
        LINK_WITH_KW.evolve(since='1.2.0'),
        LINK_WHOLE_KW.evolve(since='1.8.0'),
        *RUST_TEST_KWS,
    )
    def test(self, state: ModuleState, args: T.Tuple[str, BuildTarget], kwargs: FuncTest) -> ModuleReturnValue:
        name, _ = args
        new_target, tkwargs = self.test_common('test', state, args, kwargs)
        test: Test = self.interpreter.make_test(
            self.interpreter.current_node, (name, new_target), tkwargs)

        return ModuleReturnValue(None, [new_target, test])

    @FeatureNew('rust.doctest', '1.8.0')
    @typed_pos_args('rust.doctest', str, BuildTarget)
    @typed_kwargs(
        'rust.doctest',
        *TEST_KWS_NO_ARGS,
        DEPENDENCIES_KW,
        LINK_WITH_KW,
        LINK_WHOLE_KW,
        *RUST_TEST_KWS,
        KwargInfo(
            'args',
            ContainerTypeInfo(list, str),
            listify=True,
            default=[],
            validator=no_spaces_validator,
        ),
    )
    def doctest(self, state: ModuleState, args: T.Tuple[str, T.Union[SharedLibrary, StaticLibrary]], kwargs: FuncDoctest) -> ModuleReturnValue:
        name, base_target = args

        if not base_target.uses_rust():
            raise MesonException('doc tests are only supported for Rust targets')
        if not base_target.uses_rust_abi():
            raise MesonException("doc tests are not supported for rust_abi: 'c'")
        if state.environment.is_cross_build() and state.environment.need_exe_wrapper(base_target.for_machine):
            mlog.notice('skipping Rust doctests due to cross compilation', once=True)
            return ModuleReturnValue(None, [])

        # Link the base target's crate into the tests
        kwargs['link_with'].append(base_target)
        kwargs['depends'].append(base_target)
        workdir = kwargs['workdir']
        kwargs['workdir'] = None
        new_target, tkwargs = self.test_common('doctest', state, args, kwargs)

        # added automatically by rustdoc; keep things simple
        tkwargs['args'].remove('--test')

        # --test-args= is "parsed" simply via the Rust function split_whitespace().
        # This means no quoting nightmares (pfew) but it also means no spaces.
        # Unfortunately it's pretty hard at this point to accept e.g. CustomTarget,
        # because their paths may not be known.  This is not a big deal because the
        # user does not control the test harness, so make things easy and allow
        # strings only.
        if tkwargs['args']:
            tkwargs['args'] = ['--test-args=' + ' '.join(T.cast('T.Sequence[str]', tkwargs['args']))]
        if workdir:
            tkwargs['args'].append('--test-run-directory=' + workdir)

        if self.rustdoc[base_target.for_machine] is None:
            rustc = T.cast('RustCompiler', base_target.compilers['rust'])
            rustdoc = rustc.get_rustdoc(state.environment)
            if rustdoc:
                self.rustdoc[base_target.for_machine] = ExternalProgram(rustdoc.get_exe())
            else:
                self.rustdoc[base_target.for_machine] = NonExistingExternalProgram()

        rustdoc_prog = self.rustdoc[base_target.for_machine]
        if not rustdoc_prog.found():
            raise MesonException(f'could not find rustdoc for {base_target.for_machine} machine')

        doctests: Doctest = self.interpreter.make_test(
            self.interpreter.current_node, (name, rustdoc_prog), tkwargs, Doctest)

        # Note that the new_target is intentionally not returned, as it
        # is only reached via the base_target and never built by "ninja"
        doctests.target = new_target
        base_target.doctests = doctests
        return ModuleReturnValue(None, [doctests])

    @noPosargs
    @typed_kwargs(
        'rust.bindgen',
        KwargInfo('c_args', ContainerTypeInfo(list, str), default=[], listify=True),
        KwargInfo('args', ContainerTypeInfo(list, str), default=[], listify=True),
        KwargInfo(
            'input',
            ContainerTypeInfo(list, (File, GeneratedList, BuildTarget, BothLibraries, ExtractedObjects, CustomTargetIndex, CustomTarget, str), allow_empty=False),
            default=[],
            listify=True,
            required=True,
        ),
        KwargInfo('language', (str, NoneType), since='1.4.0', validator=in_set_validator({'c', 'cpp'})),
        KwargInfo('bindgen_version', ContainerTypeInfo(list, str), default=[], listify=True, since='1.4.0'),
        INCLUDE_DIRECTORIES.evolve(since_values={ContainerTypeInfo(list, str): '1.0.0'}),
        OUTPUT_KW,
        KwargInfo(
            'output_inline_wrapper',
            str,
            default='',
            since='1.4.0',
        ),
        DEPENDENCIES_KW.evolve(since='1.0.0'),
    )
    def bindgen(self, state: ModuleState, args: T.List, kwargs: FuncBindgen) -> ModuleReturnValue:
        """Wrapper around bindgen to simplify its use.

        The main thing this simplifies is the use of `include_directory`
        objects, instead of having to pass a plethora of `-I` arguments.
        """
        header, *_deps = self.interpreter.source_strings_to_files(kwargs['input'])

        # Split File and Target dependencies to add pass to CustomTarget
        depends: T.List[SourceOutputs] = []
        depend_files: T.List[File] = []
        for d in _deps:
            if isinstance(d, File):
                depend_files.append(d)
            else:
                depends.append(d)

        # Copy to avoid subsequent calls mutating the original
        # TODO: if we want this to be per-machine we'll need a native kwarg
        clang_args = state.environment.properties.host.get_bindgen_clang_args().copy()

        # Find the first C'ish compiler to fetch the default compiler flags
        # from. Append those to the bindgen flags to ensure we use a compatible
        # environment.
        comp = mesonlib.first(
            [state.environment.coredata.compilers.host.get(l) for l in ['c', 'cpp', 'objc', 'objcpp']],
            lambda x: x is not None,
        )
        if comp:
            clang_args.extend(comp.get_always_args())
        else:
            mlog.warning(textwrap.dedent('''\
                Using `rust.bindgen` without configuring C (or a C-like)
                language in Meson will skip compiler detection and can cause
                ABI incompatibilities due to missing crucial compiler flags.
                Consider calling `add_languages('c')` in your Meson build
                files.
            '''))

        for i in state.process_include_dirs(kwargs['include_directories']):
            # bindgen always uses clang, so it's safe to hardcode -I here
            clang_args.extend([f'-I{x}' for x in i.to_string_list(
                state.environment.get_source_dir(), state.environment.get_build_dir())])
        if are_asserts_disabled_for_subproject(state.subproject, state.environment):
            clang_args.append('-DNDEBUG')

        for de in kwargs['dependencies']:
            for i in de.get_include_dirs():
                clang_args.extend([f'-I{x}' for x in i.to_string_list(
                    state.environment.get_source_dir(), state.environment.get_build_dir())])
            clang_args.extend(de.get_all_compile_args())
            for s in de.get_sources():
                if isinstance(s, File):
                    depend_files.append(s)
                elif isinstance(s, CustomTarget):
                    depends.append(s)

        if self._bindgen_bin is None:
            self._bindgen_bin = state.find_program('bindgen', wanted=kwargs['bindgen_version'])
            if self._bindgen_rust_target is not None:
                # ExternalCommand.command's type is bonkers
                _, _, err = mesonlib.Popen_safe(
                    T.cast('T.List[str]', self._bindgen_bin.get_command()) +
                    ['--rust-target', self._bindgen_rust_target])
                # < 0.71: Sometimes this is "invalid Rust target" and
                # sometimes "invalid # rust target"
                # >= 0.71: error: invalid value '...' for '--rust-target <RUST_TARGET>': "..." is not a valid Rust target, accepted values are of the form ...
                # It's also much harder to hit this in 0.71 than in previous versions
                if 'Got an invalid' in err or 'is not a valid Rust target' in err:
                    self._bindgen_rust_target = None

            # TODO: Executable needs to learn about get_version
            if isinstance(self._bindgen_bin, ExternalProgram):
                self._bindgen_set_std = mesonlib.version_compare(self._bindgen_bin.get_version(), '>= 0.71')

        name: str
        if isinstance(header, File):
            name = header.fname
        elif isinstance(header, (BuildTarget, BothLibraries, ExtractedObjects, StructuredSources)):
            raise InterpreterException('bindgen source file must be a C header, not an object or build target')
        else:
            name = header.get_outputs()[0]

        # bindgen assumes that C++ headers will be called .hpp. We want to
        # ensure that anything Meson considers a C++ header is treated as one.
        language = kwargs['language']
        if language is None:
            ext = os.path.splitext(name)[1][1:]
            if ext in lang_suffixes['cpp']:
                language = 'cpp'
            elif ext == 'h':
                language = 'c'
            else:
                raise InterpreterException(f'Unknown file type extension for: {name}')

        # We only want include directories and defines, other things may not be valid
        cargs = state.get_option(f'{language}_args', state.subproject)
        assert isinstance(cargs, list), 'for mypy'
        for a in itertools.chain(state.global_args.get(language, []), state.project_args.get(language, []), cargs):
            if a.startswith(('-I', '/I', '-D', '/D', '-U', '/U')):
                clang_args.append(a)

        if language == 'cpp':
            clang_args.extend(['-x', 'c++'])

        # Add the C++ standard to the clang arguments. Attempt to translate VS
        # extension versions into the nearest standard version
        std = state.get_option(f'{language}_std')
        assert isinstance(std, str), 'for mypy'
        if std.startswith('vc++'):
            if std.endswith('latest'):
                mlog.warning('Attempting to translate vc++latest into a clang compatible version.',
                             'Currently this is hardcoded for c++20', once=True, fatal=False)
                std = 'c++20'
            else:
                mlog.debug('The current C++ standard is a Visual Studio extension version.',
                           'bindgen will use a the nearest C++ standard instead')
                std = std[1:]

        if std != 'none':
            clang_args.append(f'-std={std}')

        inline_wrapper_args: T.List[str] = []
        outputs = [kwargs['output']]
        if kwargs['output_inline_wrapper']:
            # Todo drop this isinstance once Executable supports version_compare
            if isinstance(self._bindgen_bin, ExternalProgram):
                if mesonlib.version_compare(self._bindgen_bin.get_version(), '< 0.65'):
                    raise InterpreterException('\'output_inline_wrapper\' parameter of rust.bindgen requires bindgen-0.65 or newer')

            outputs.append(kwargs['output_inline_wrapper'])
            inline_wrapper_args = [
                '--experimental', '--wrap-static-fns',
                '--wrap-static-fns-path', os.path.join(state.environment.build_dir, '@OUTPUT1@')
            ]

        cmd = self._bindgen_bin.get_command() + \
            [
                '@INPUT@', '--output',
                os.path.join(state.environment.build_dir, '@OUTPUT0@')
            ] + \
            kwargs['args'] + inline_wrapper_args
        if self._bindgen_rust_target and '--rust-target' not in cmd:
            cmd.extend(['--rust-target', self._bindgen_rust_target])
        if self._bindgen_set_std and '--rust-edition' not in cmd:
            rust_std = state.environment.coredata.optstore.get_value('rust_std')
            assert isinstance(rust_std, str), 'for mypy'
            if rust_std != 'none':
                cmd.extend(['--rust-edition', rust_std])
        cmd.append('--')
        cmd.extend(kwargs['c_args'])
        cmd.extend(clang_args)
        cmd.extend(['-MD', '-MQ', '@INPUT@', '-MF', '@DEPFILE@'])

        target = CustomTarget(
            f'rustmod-bindgen-{name}'.replace('/', '_'),
            state.subdir,
            state.subproject,
            state.environment,
            cmd,
            [header],
            outputs,
            depfile='@PLAINNAME@.d',
            extra_depends=depends,
            depend_files=depend_files,
            backend=state.backend,
            description='Generating bindings for Rust {}',
        )

        return ModuleReturnValue(target, [target])

    # Allow a limited set of kwargs, but still use the full set of typed_kwargs()
    # because it could be setting required default values.
    @FeatureNew('rust.proc_macro', '1.3.0')
    @permittedKwargs({'rust_args', 'rust_dependency_map', 'sources', 'dependencies', 'extra_files',
                      'link_args', 'link_depends', 'link_with', 'override_options'})
    @typed_pos_args('rust.proc_macro', str, varargs=SOURCES_VARARGS)
    @typed_kwargs('rust.proc_macro', *SHARED_LIB_KWS, allow_unknown=True)
    def proc_macro(self, state: ModuleState, args: T.Tuple[str, SourcesVarargsType], kwargs: _kwargs.SharedLibrary) -> SharedLibrary:
        kwargs['native'] = True  # type: ignore
        kwargs['rust_crate_type'] = 'proc-macro'  # type: ignore
        kwargs['rust_args'] = kwargs['rust_args'] + ['--extern', 'proc_macro']
        target = state._interpreter.build_target(state.current_node, args, kwargs, SharedLibrary)
        return target


def initialize(interp: Interpreter) -> RustModule:
    return RustModule(interp)



# SPDX-License-Identifier: Apache-2.0
# Copyright 2014-2016 The Meson development team

from __future__ import annotations

"""This is a helper script for IDE developers. It allows you to
extract information such as list of targets, files, compiler flags,
tests and so on. All output is in JSON for simple parsing.

Currently only works for the Ninja backend. Others use generated
project files and don't need this info."""

from contextlib import redirect_stdout
import collections
import dataclasses
import json
import os
from pathlib import Path, PurePath
import sys
import typing as T

from . import build, environment, mesonlib, options, coredata as cdata
from .ast import IntrospectionInterpreter, AstConditionLevel, AstIDGenerator, AstIndentationGenerator, AstJSONPrinter
from .backend import backends
from .dependencies import Dependency
from .interpreterbase import ObjectHolder, UnknownValue
from .options import OptionKey

if T.TYPE_CHECKING:
    import argparse

    from .interpreter import Interpreter

class IntrospectionEncoder(json.JSONEncoder):
    def default(self, obj: T.Any) -> T.Any:
        if isinstance(obj, UnknownValue):
            return 'unknown'
        return json.JSONEncoder.default(self, obj)

def get_meson_info_file(info_dir: str) -> str:
    return os.path.join(info_dir, 'meson-info.json')

def get_meson_introspection_version() -> str:
    return '1.0.0'

def get_meson_introspection_required_version() -> T.List[str]:
    return ['>=1.0', '<2.0']

class IntroCommand:
    def __init__(self,
                 desc: str,
                 func: T.Optional[T.Callable[[], T.Union[dict, list]]] = None,
                 no_bd: T.Optional[T.Callable[[IntrospectionInterpreter], T.Union[dict, list]]] = None) -> None:
        self.desc = desc + '.'
        self.func = func
        self.no_bd = no_bd

def get_meson_introspection_types(coredata: T.Optional[cdata.CoreData] = None,
                                  builddata: T.Optional[build.Build] = None,
                                  backend: T.Optional[backends.Backend] = None) -> T.Mapping[str, IntroCommand]:
    if backend and builddata:
        benchmarkdata = backend.create_test_serialisation(builddata.get_benchmarks())
        testdata = backend.create_test_serialisation(builddata.get_tests())
        installdata = backend.create_install_data()
        interpreter = backend.interpreter
    else:
        benchmarkdata = testdata = installdata = None

    # Enforce key order for argparse
    return collections.OrderedDict([
        ('ast', IntroCommand('Dump the AST of the meson file', no_bd=dump_ast)),
        ('benchmarks', IntroCommand('List all benchmarks', func=lambda: list_benchmarks(benchmarkdata))),
        ('buildoptions', IntroCommand('List all build options', func=lambda: list_buildoptions(coredata), no_bd=list_buildoptions_from_source)),
        ('buildsystem_files', IntroCommand('List files that make up the build system', func=lambda: list_buildsystem_files(builddata, interpreter))),
        ('compilers', IntroCommand('List used compilers', func=lambda: list_compilers(coredata))),
        ('dependencies', IntroCommand('List external dependencies', func=lambda: list_deps(coredata, backend), no_bd=list_deps_from_source)),
        ('scan_dependencies', IntroCommand('Scan for dependencies used in the meson.build file', no_bd=list_deps_from_source)),
        ('installed', IntroCommand('List all installed files and directories', func=lambda: list_installed(installdata))),
        ('install_plan', IntroCommand('List all installed files and directories with their details', func=lambda: list_install_plan(installdata))),
        ('machines', IntroCommand('Information about host, build, and target machines', func=lambda: list_machines(builddata))),
        ('projectinfo', IntroCommand('Information about projects', func=lambda: list_projinfo(builddata), no_bd=list_projinfo_from_source)),
        ('targets', IntroCommand('List top level targets', func=lambda: list_targets(builddata, installdata, backend), no_bd=list_targets_from_source)),
        ('tests', IntroCommand('List all unit tests', func=lambda: list_tests(testdata))),
    ])

# Note: when adding arguments, please also add them to the completion
# scripts in $MESONSRC/data/shell-completions/
def add_arguments(parser: argparse.ArgumentParser) -> None:
    intro_types = get_meson_introspection_types()
    for key, val in intro_types.items():
        flag = '--' + key.replace('_', '-')
        parser.add_argument(flag, action='store_true', dest=key, default=False, help=val.desc)

    parser.add_argument('--backend', choices=sorted(options.backendlist), dest='backend', default='ninja',
                        help='The backend to use for the --buildoptions introspection.')
    parser.add_argument('-a', '--all', action='store_true', dest='all', default=False,
                        help='Print all available information.')
    parser.add_argument('-i', '--indent', action='store_true', dest='indent', default=False,
                        help='Enable pretty printed JSON.')
    parser.add_argument('-f', '--force-object-output', action='store_true', dest='force_dict', default=False,
                        help='Always use the new JSON format for multiple entries (even for 0 and 1 introspection commands)')
    parser.add_argument('builddir', nargs='?', default='.', help='The build directory')

def dump_ast(intr: IntrospectionInterpreter) -> T.Dict[str, T.Any]:
    printer = AstJSONPrinter()
    intr.ast.accept(printer)
    return printer.result

def list_installed(installdata: backends.InstallData) -> T.Dict[str, str]:
    res = {}
    if installdata is not None:
        for t in installdata.targets:
            res[os.path.join(installdata.build_dir, t.fname)] = \
                os.path.join(installdata.prefix, t.outdir, os.path.basename(t.fname))
        for i in installdata.data:
            res[i.path] = os.path.join(installdata.prefix, i.install_path)
        for i in installdata.headers:
            res[i.path] = os.path.join(installdata.prefix, i.install_path, os.path.basename(i.path))
        for i in installdata.man:
            res[i.path] = os.path.join(installdata.prefix, i.install_path)
        for i in installdata.install_subdirs:
            res[i.path] = os.path.join(installdata.prefix, i.install_path)
        for s in installdata.symlinks:
            basename = os.path.basename(s.name)
            res[basename] = os.path.join(installdata.prefix, s.install_path, basename)
    return res

def list_install_plan(installdata: backends.InstallData) -> T.Dict[str, T.Dict[str, T.Dict[str, T.Union[str, T.List[str], None]]]]:
    plan: T.Dict[str, T.Dict[str, T.Dict[str, T.Union[str, T.List[str], None]]]] = {
        'targets': {
            os.path.join(installdata.build_dir, target.fname): {
                'destination': target.out_name,
                'tag': target.tag or None,
                'subproject': target.subproject or None,
                'install_rpath': target.install_rpath or None,
                'build_rpaths': sorted(x.decode('utf8') for x in target.rpath_dirs_to_remove),
            }
            for target in installdata.targets
        },
    }
    for key, data_list in {
        'data': installdata.data,
        'man': installdata.man,
        'headers': installdata.headers,
        'install_subdirs': installdata.install_subdirs
    }.items():
        # Mypy doesn't recognize SubdirInstallData as a subclass of InstallDataBase
        for data in data_list: # type: ignore[attr-defined]
            data_type = data.data_type or key
            install_path_name = data.install_path_name
            if key == 'headers':  # in the headers, install_path_name is the directory
                install_path_name = os.path.join(install_path_name, os.path.basename(data.path))

            entry = {
                'destination': install_path_name,
                'tag': data.tag or None,
                'subproject': data.subproject or None,
            }

            if key == 'install_subdirs':
                exclude_files, exclude_dirs = data.exclude or ([], [])
                entry['exclude_dirs'] = list(exclude_dirs)
                entry['exclude_files'] = list(exclude_files)

            plan[data_type] = plan.get(data_type, {})
            plan[data_type][data.path] = entry

    return plan

def get_target_dir(coredata: cdata.CoreData, subdir: str) -> str:
    if coredata.optstore.get_value_for(OptionKey('layout')) == 'flat':
        return 'meson-out'
    else:
        return subdir

def list_targets_from_source(intr: IntrospectionInterpreter) -> T.List[T.Dict[str, object]]:
    tlist = []
    root_dir = Path(intr.source_root).resolve()

    for i in intr.targets:
        sources = intr.nodes_to_pretty_filelist(root_dir, i.subdir, i.source_nodes)
        extra_files = intr.nodes_to_pretty_filelist(root_dir, i.subdir, [i.extra_files] if i.extra_files else [])

        outdir = get_target_dir(intr.coredata, i.subdir)

        tlist += [{
            'name': i.name,
            'id': i.id,
            'type': i.typename,
            'defined_in': i.defined_in,
            'filename': [os.path.join(outdir, x) for x in i.outputs],
            'build_by_default': i.build_by_default,
            'target_sources': [{
                'language': 'unknown',
                'machine': i.machine,
                'compiler': [],
                'parameters': [],
                'sources': sources,
                'generated_sources': []
            }],
            'depends': [],
            'extra_files': extra_files,
            'subproject': None, # Subprojects are not supported
            'installed': i.installed
        }]

    return tlist

def list_targets(builddata: build.Build, installdata: backends.InstallData, backend: backends.Backend) -> T.List[T.Any]:
    tlist: T.List[T.Any] = []
    build_dir = builddata.environment.get_build_dir()
    src_dir = builddata.environment.get_source_dir()

    # Fast lookup table for installation files
    install_lookuptable = {}
    for i in installdata.targets:
        basename = os.path.basename(i.fname)
        install_lookuptable[basename] = [str(PurePath(installdata.prefix, i.outdir, basename))]
    for s in installdata.symlinks:
        # Symlink's target must already be in the table. They share the same list
        # to support symlinks to symlinks recursively, such as .so -> .so.0 -> .so.1.2.3
        basename = os.path.basename(s.name)
        try:
            install_lookuptable[basename] = install_lookuptable[os.path.basename(s.target)]
            install_lookuptable[basename].append(str(PurePath(installdata.prefix, s.install_path, basename)))
        except KeyError:
            pass

    for (idname, target) in builddata.get_targets().items():
        if not isinstance(target, build.Target):
            raise RuntimeError('The target object in `builddata.get_targets()` is not of type `build.Target`. Please file a bug with this error message.')

        outdir = get_target_dir(builddata.environment.coredata, target.subdir)
        t = {
            'name': target.get_basename(),
            'id': idname,
            'type': target.get_typename(),
            'defined_in': os.path.normpath(os.path.join(src_dir, target.subdir, environment.build_filename)),
            'filename': [os.path.join(build_dir, outdir, x) for x in target.get_outputs()],
            'build_by_default': target.build_by_default,
            'target_sources': backend.get_introspection_data(idname, target),
            'extra_files': [os.path.normpath(os.path.join(src_dir, x.subdir, x.fname)) for x in target.extra_files],
            'subproject': target.subproject or None,
            'dependencies': [d.name for d in getattr(target, 'external_deps', [])],
            'depends': [lib.get_id() for lib in getattr(target, 'dependencies', [])]
        }

        vs_module_defs = getattr(target, 'vs_module_defs', None)
        if vs_module_defs is not None:
            t['vs_module_defs'] = vs_module_defs.relative_name()
        win_subsystem = getattr(target, 'win_subsystem', None)
        if win_subsystem is not None:
            t['win_subsystem'] = win_subsystem

        if installdata and target.should_install():
            t['installed'] = True
            ifn = [install_lookuptable.get(x, [None]) for x in target.get_outputs()]
            t['install_filename'] = [x for sublist in ifn for x in sublist]  # flatten the list
        else:
            t['installed'] = False
        tlist.append(t)
    return tlist

def list_buildoptions_from_source(intr: IntrospectionInterpreter) -> T.List[T.Dict[str, T.Union[str, bool, int, T.List[str]]]]:
    subprojects = [i['name'] for i in intr.project_data['subprojects']]
    return list_buildoptions(intr.coredata, subprojects)

def list_buildoptions(coredata: cdata.CoreData, subprojects: T.Optional[T.List[str]] = None) -> T.List[T.Dict[str, T.Union[str, bool, int, T.List[str]]]]:
    optlist: T.List[T.Dict[str, T.Union[str, bool, int, T.List[str]]]] = []
    subprojects = subprojects or []

    dir_option_names = set(options.BUILTIN_DIR_OPTIONS)
    test_option_names = {OptionKey('errorlogs'),
                         OptionKey('stdsplit')}

    dir_options: options.MutableKeyedOptionDictType = {}
    test_options: options.MutableKeyedOptionDictType = {}
    core_options: options.MutableKeyedOptionDictType = {}
    for k, v in coredata.optstore.items():
        if k in dir_option_names:
            dir_options[k] = v
        elif k in test_option_names:
            test_options[k] = v
        elif coredata.optstore.is_builtin_option(k):
            core_options[k] = v
            if not v.yielding:
                for s in subprojects:
                    core_options[k.evolve(subproject=s)] = v

    def add_keys(opts: T.Union[options.MutableKeyedOptionDictType, options.OptionStore], section: str) -> None:
        for key, opt in sorted(opts.items()):
            optdict = {'name': str(key), 'value': opt.value, 'section': section,
                       'machine': key.machine.get_lower_case_name() if coredata.optstore.is_per_machine_option(key) else 'any'}
            if isinstance(opt, options.UserStringOption):
                typestr = 'string'
            elif isinstance(opt, options.UserBooleanOption):
                typestr = 'boolean'
            elif isinstance(opt, options.UserComboOption):
                optdict['choices'] = opt.printable_choices()
                typestr = 'combo'
            elif isinstance(opt, (options.UserIntegerOption, options.UserUmaskOption)):
                typestr = 'integer'
            elif isinstance(opt, options.UserStringArrayOption):
                typestr = 'array'
                c = opt.printable_choices()
                if c:
                    optdict['choices'] = c
            else:
                raise RuntimeError('Unknown option type: ', repr(type(opt)))
            optdict['type'] = typestr
            optdict['description'] = opt.description
            optlist.append(optdict)

    add_keys(core_options, 'core')
    add_keys({k: v for k, v in coredata.optstore.items() if coredata.optstore.is_backend_option(k)}, 'backend')
    add_keys({k: v for k, v in coredata.optstore.items() if coredata.optstore.is_base_option(k)}, 'base')
    add_keys(
        {k: v for k, v in sorted(coredata.optstore.items(), key=lambda i: i[0].machine) if coredata.optstore.is_compiler_option(k)},
        'compiler',
    )
    add_keys(dir_options, 'directory')

    def project_option_key_to_introname(key: OptionKey) -> OptionKey:
        assert key.subproject is not None
        if key.subproject == '':
            return key.evolve(subproject=None)
        return key

    add_keys({project_option_key_to_introname(k): v
              for k, v in coredata.optstore.items() if coredata.optstore.is_project_option(k)}, 'user')
    add_keys(test_options, 'test')
    return optlist

def find_buildsystem_files_list(src_dir: str) -> T.List[str]:
    build_files = frozenset({'meson.build', 'meson.options', 'meson_options.txt'})
    # I feel dirty about this. But only slightly.
    filelist: T.List[str] = []
    for root, _, files in os.walk(src_dir):
        filelist.extend(os.path.relpath(os.path.join(root, f), src_dir)
                        for f in build_files.intersection(files))
    return filelist

def list_buildsystem_files(builddata: build.Build, interpreter: Interpreter) -> T.List[str]:
    src_dir = builddata.environment.get_source_dir()
    filelist = list(interpreter.get_build_def_files())
    filelist = [PurePath(src_dir, x).as_posix() for x in filelist]
    return filelist

def list_compilers(coredata: cdata.CoreData) -> T.Dict[str, T.Dict[str, T.Dict[str, str]]]:
    compilers: T.Dict[str, T.Dict[str, T.Dict[str, str]]] = {}
    for machine in ('host', 'build'):
        compilers[machine] = {}
        for language, compiler in getattr(coredata.compilers, machine).items():
            compilers[machine][language] = {
                'id': compiler.get_id(),
                'exelist': compiler.get_exelist(),
                'linker_exelist': compiler.get_linker_exelist(),
                'file_suffixes': compiler.file_suffixes,
                'default_suffix': compiler.get_default_suffix(),
                'version': compiler.version,
                'full_version': compiler.full_version,
                'linker_id': compiler.get_linker_id(),
            }
    return compilers

def list_deps_from_source(intr: IntrospectionInterpreter) -> T.List[T.Dict[str, T.Union[str, bool, T.List[str], UnknownValue]]]:
    result: T.List[T.Dict[str, T.Union[str, bool, T.List[str], UnknownValue]]] = []
    for i in intr.dependencies:
        result += [{
            'name': i.name,
            'required': i.required,
            'version': i.version,
            'has_fallback': i.has_fallback,
            'conditional': i.conditional,
        }]
    return result

def list_deps(coredata: cdata.CoreData, backend: backends.Backend) -> T.List[T.Dict[str, T.Union[str, T.List[str]]]]:
    result: T.Dict[str, T.Dict[str, T.Union[str, T.List[str]]]] = {}

    def _src_to_str(src_file: T.Union[mesonlib.FileOrString, build.CustomTarget, build.StructuredSources, build.CustomTargetIndex, build.GeneratedList]) -> T.List[str]:
        if isinstance(src_file, str):
            return [src_file]
        if isinstance(src_file, mesonlib.File):
            return [src_file.absolute_path(backend.source_dir, backend.build_dir)]
        if isinstance(src_file, (build.CustomTarget, build.CustomTargetIndex, build.GeneratedList)):
            return src_file.get_outputs()
        if isinstance(src_file, build.StructuredSources):
            return [f for s in src_file.as_list() for f in _src_to_str(s)]
        raise mesonlib.MesonBugException(f'Invalid file type {type(src_file)}.')

    def _create_result(d: Dependency, varname: T.Optional[str] = None) -> T.Dict[str, T.Any]:
        return {
            'name': d.name,
            'type': d.type_name,
            'version': d.get_version(),
            'compile_args': d.get_compile_args(),
            'link_args': d.get_link_args(),
            'include_directories': [i for idirs in d.get_include_dirs() for i in idirs.to_string_list(backend.source_dir, backend.build_dir)],
            'sources': [f for s in d.get_sources() for f in _src_to_str(s)],
            'extra_files': [f for s in d.get_extra_files() for f in _src_to_str(s)],
            'dependencies': [e.name for e in d.ext_deps],
            'depends': [lib.get_id() for lib in getattr(d, 'libraries', [])],
            'meson_variables': [varname] if varname else [],
        }

    for d in coredata.deps.host.values():
        if d.found():
            result[d.name] = _create_result(d)

    for varname, holder in backend.interpreter.variables.items():
        if isinstance(holder, ObjectHolder):
            d = holder.held_object
            if isinstance(d, Dependency) and d.found():
                if d.name in result:
                    T.cast('T.List[str]', result[d.name]['meson_variables']).append(varname)
                else:
                    result[d.name] = _create_result(d, varname)

    return list(result.values())

def get_test_list(testdata: T.List[backends.TestSerialisation]) -> T.List[T.Dict[str, T.Union[str, int, T.List[str], T.Dict[str, str]]]]:
    result: T.List[T.Dict[str, T.Union[str, int, T.List[str], T.Dict[str, str]]]] = []
    for t in testdata:
        to: T.Dict[str, T.Union[str, int, T.List[str], T.Dict[str, str]]] = {}
        if isinstance(t.fname, str):
            fname = [t.fname]
        else:
            fname = t.fname
        to['cmd'] = fname + t.cmd_args
        if isinstance(t.env, mesonlib.EnvironmentVariables):
            to['env'] = t.env.get_env({})
        else:
            to['env'] = t.env
        to['name'] = t.name
        to['workdir'] = t.workdir
        to['timeout'] = t.timeout
        to['suite'] = t.suite
        to['is_parallel'] = t.is_parallel
        to['priority'] = t.priority
        to['protocol'] = str(t.protocol)
        to['depends'] = t.depends
        to['extra_paths'] = t.extra_paths
        result.append(to)
    return result

def list_tests(testdata: T.List[backends.TestSerialisation]) -> T.List[T.Dict[str, T.Union[str, int, T.List[str], T.Dict[str, str]]]]:
    return get_test_list(testdata)

def list_benchmarks(benchdata: T.List[backends.TestSerialisation]) -> T.List[T.Dict[str, T.Union[str, int, T.List[str], T.Dict[str, str]]]]:
    return get_test_list(benchdata)

def list_machines(builddata: build.Build) -> T.Dict[str, T.Dict[str, T.Union[str, bool]]]:
    machines: T.Dict[str, T.Dict[str, T.Union[str, bool]]] = {}
    for m in ('host', 'build', 'target'):
        machine = getattr(builddata.environment.machines, m)
        machines[m] = dataclasses.asdict(machine)
        machines[m]['is_64_bit'] = machine.is_64_bit
        machines[m]['exe_suffix'] = machine.get_exe_suffix()
        machines[m]['object_suffix'] = machine.get_object_suffix()
    return machines

def list_projinfo(builddata: build.Build) -> T.Dict[str, T.Union[str, T.List[str], T.List[T.Dict[str, str]]]]:
    result: T.Dict[str, T.Union[str, T.List[str], T.List[T.Dict[str, str]]]] = {
        'version': builddata.project_version,
        'descriptive_name': builddata.project_name,
        'license': builddata.dep_manifest[builddata.project_name].license,
        'license_files': [f[1].fname for f in builddata.dep_manifest[builddata.project_name].license_files],
        'subproject_dir': builddata.subproject_dir,
    }
    subprojects = []
    for k, v in builddata.subprojects.items():
        c: T.Dict[str, str] = {
            'name': k,
            'version': v,
            'descriptive_name': builddata.projects.get(k),
        }
        subprojects.append(c)
    result['subprojects'] = subprojects
    return result

def list_projinfo_from_source(intr: IntrospectionInterpreter) -> T.Dict[str, T.Union[str, T.List[T.Dict[str, str]]]]:
    sourcedir = intr.source_root
    files = find_buildsystem_files_list(sourcedir)
    files = [os.path.normpath(x) for x in files]

    for i in intr.project_data['subprojects']:
        basedir = os.path.join(intr.subproject_dir, i['name'])
        i['buildsystem_files'] = [x for x in files if x.startswith(basedir)]
        files = [x for x in files if not x.startswith(basedir)]

    intr.project_data['buildsystem_files'] = files
    intr.project_data['subproject_dir'] = intr.subproject_dir
    return intr.project_data

def print_results(options: argparse.Namespace, results: T.Sequence[T.Tuple[str, T.Union[dict, T.List[T.Any]]]], indent: T.Optional[int]) -> int:
    if not results and not options.force_dict:
        print('No command specified')
        return 1
    elif len(results) == 1 and not options.force_dict:
        # Make to keep the existing output format for a single option
        print(json.dumps(results[0][1], indent=indent, cls=IntrospectionEncoder))
    else:
        out = {}
        for i in results:
            out[i[0]] = i[1]
        print(json.dumps(out, indent=indent, cls=IntrospectionEncoder))
    return 0

def get_infodir(builddir: T.Optional[str] = None) -> str:
    infodir = 'meson-info'
    if builddir is not None:
        infodir = os.path.join(builddir, infodir)
    return infodir

def get_info_file(infodir: str, kind: T.Optional[str] = None) -> str:
    return os.path.join(infodir,
                        'meson-info.json' if not kind else f'intro-{kind}.json')

def load_info_file(infodir: str, kind: T.Optional[str] = None) -> T.Any:
    with open(get_info_file(infodir, kind), encoding='utf-8') as fp:
        return json.load(fp)

def run(options: argparse.Namespace) -> int:
    datadir = 'meson-private'
    infodir = get_infodir(options.builddir)
    if options.builddir is not None:
        datadir = os.path.join(options.builddir, datadir)
    indent = 4 if options.indent else None
    results: T.List[T.Tuple[str, T.Union[dict, T.List[T.Any]]]] = []
    intro_types = get_meson_introspection_types()

    # TODO: This if clause is undocumented.
    if os.path.basename(options.builddir) == environment.build_filename:
        sourcedir = '.' if options.builddir == environment.build_filename else options.builddir[:-len(environment.build_filename)]
        # Make sure that log entries in other parts of meson don't interfere with the JSON output
        with redirect_stdout(sys.stderr):
            backend = backends.get_backend_from_name(options.backend)
            assert backend is not None
            intr = IntrospectionInterpreter(sourcedir, '', backend.name, visitors = [AstIDGenerator(), AstIndentationGenerator(), AstConditionLevel()])
            intr.analyze()

        for key, val in intro_types.items():
            if (not options.all and not getattr(options, key, False)) or not val.no_bd:
                continue
            results += [(key, val.no_bd(intr))]
        return print_results(options, results, indent)

    try:
        raw = load_info_file(infodir)
        intro_vers = raw.get('introspection', {}).get('version', {}).get('full', '0.0.0')
    except FileNotFoundError:
        if not os.path.isdir(datadir) or not os.path.isdir(infodir):
            print('Current directory is not a meson build directory.\n'
                  'Please specify a valid build dir or change the working directory to it.')
        else:
            print('Introspection file {} does not exist.\n'
                  'It is also possible that the build directory was generated with an old\n'
                  'meson version. Please regenerate it in this case.'.format(get_info_file(infodir)))
        return 1

    vers_to_check = get_meson_introspection_required_version()
    for i in vers_to_check:
        if not mesonlib.version_compare(intro_vers, i):
            print('Introspection version {} is not supported. '
                  'The required version is: {}'
                  .format(intro_vers, ' and '.join(vers_to_check)))
            return 1

    # Extract introspection information from JSON
    for i, v in intro_types.items():
        if not v.func:
            continue
        if not options.all and not getattr(options, i, False):
            continue
        try:
            results += [(i, load_info_file(infodir, i))]
        except FileNotFoundError:
            print('Introspection file {} does not exist.'.format(get_info_file(infodir, i)))
            return 1

    return print_results(options, results, indent)

updated_introspection_files: T.List[str] = []

def write_intro_info(intro_info: T.Sequence[T.Tuple[str, T.Union[dict, T.List[T.Any]]]], info_dir: str) -> None:
    for kind, data in intro_info:
        out_file = os.path.join(info_dir, f'intro-{kind}.json')
        tmp_file = os.path.join(info_dir, 'tmp_dump.json')
        with open(tmp_file, 'w', encoding='utf-8') as fp:
            json.dump(data, fp, indent=2)
            fp.flush() # Not sure if this is needed
        os.replace(tmp_file, out_file)
        updated_introspection_files.append(kind)

def generate_introspection_file(builddata: build.Build, backend: backends.Backend) -> None:
    coredata = builddata.environment.get_coredata()
    intro_types = get_meson_introspection_types(coredata=coredata, builddata=builddata, backend=backend)
    intro_info: T.List[T.Tuple[str, T.Union[dict, T.List[T.Any]]]] = []

    for key, val in intro_types.items():
        if not val.func:
            continue
        intro_info += [(key, val.func())]

    write_intro_info(intro_info, builddata.environment.info_dir)

def update_build_options(coredata: cdata.CoreData, info_dir: str) -> None:
    intro_info = [
        ('buildoptions', list_buildoptions(coredata))
    ]

    write_intro_info(intro_info, info_dir)

def split_version_string(version: str) -> T.Dict[str, T.Union[str, int]]:
    vers_list = version.split('.')
    return {
        'full': version,
        'major': int(vers_list[0] if len(vers_list) > 0 else 0),
        'minor': int(vers_list[1] if len(vers_list) > 1 else 0),
        'patch': int(vers_list[2] if len(vers_list) > 2 else 0)
    }

def write_meson_info_file(builddata: build.Build, errors: list, build_files_updated: bool = False) -> None:
    info_dir = builddata.environment.info_dir
    info_file = get_meson_info_file(info_dir)
    intro_types = get_meson_introspection_types()
    intro_info = {}

    for i, v in intro_types.items():
        if not v.func:
            continue
        intro_info[i] = {
            'file': f'intro-{i}.json',
            'updated': i in updated_introspection_files
        }

    info_data = {
        'meson_version': split_version_string(cdata.version),
        'directories': {
            'source': builddata.environment.get_source_dir(),
            'build': builddata.environment.get_build_dir(),
            'info': info_dir,
        },
        'introspection': {
            'version': split_version_string(get_meson_introspection_version()),
            'information': intro_info,
        },
        'build_files_updated': build_files_updated,
    }

    if errors:
        info_data['error'] = True
        info_data['error_list'] = [x if isinstance(x, str) else str(x) for x in errors]
    else:
        info_data['error'] = False

    # Write the data to disc
    tmp_file = os.path.join(info_dir, 'tmp_dump.json')
    with open(tmp_file, 'w', encoding='utf-8') as fp:
        json.dump(info_data, fp)
        fp.flush()
    os.replace(tmp_file, info_file)



"""Implementation of JSONDecoder
"""
import re

from json import scanner
try:
    from _json import scanstring as c_scanstring
except ImportError:
    c_scanstring = None

__all__ = ['JSONDecoder', 'JSONDecodeError']

FLAGS = re.VERBOSE | re.MULTILINE | re.DOTALL

NaN = float('nan')
PosInf = float('inf')
NegInf = float('-inf')


class JSONDecodeError(ValueError):
    """Subclass of ValueError with the following additional properties:

    msg: The unformatted error message
    doc: The JSON document being parsed
    pos: The start index of doc where parsing failed
    lineno: The line corresponding to pos
    colno: The column corresponding to pos

    """
    # Note that this exception is used from _json
    def __init__(self, msg, doc, pos):
        lineno = doc.count('\n', 0, pos) + 1
        colno = pos - doc.rfind('\n', 0, pos)
        errmsg = '%s: line %d column %d (char %d)' % (msg, lineno, colno, pos)
        ValueError.__init__(self, errmsg)
        self.msg = msg
        self.doc = doc
        self.pos = pos
        self.lineno = lineno
        self.colno = colno

    def __reduce__(self):
        return self.__class__, (self.msg, self.doc, self.pos)


_CONSTANTS = {
    '-Infinity': NegInf,
    'Infinity': PosInf,
    'NaN': NaN,
}


HEXDIGITS = re.compile(r'[0-9A-Fa-f]{4}', FLAGS)
STRINGCHUNK = re.compile(r'(.*?)(["\\\x00-\x1f])', FLAGS)
BACKSLASH = {
    '"': '"', '\\': '\\', '/': '/',
    'b': '\b', 'f': '\f', 'n': '\n', 'r': '\r', 't': '\t',
}

def _decode_uXXXX(s, pos, _m=HEXDIGITS.match):
    esc = _m(s, pos + 1)
    if esc is not None:
        try:
            return int(esc.group(), 16)
        except ValueError:
            pass
    msg = "Invalid \\uXXXX escape"
    raise JSONDecodeError(msg, s, pos)

def py_scanstring(s, end, strict=True,
        _b=BACKSLASH, _m=STRINGCHUNK.match):
    """Scan the string s for a JSON string. End is the index of the
    character in s after the quote that started the JSON string.
    Unescapes all valid JSON string escape sequences and raises ValueError
    on attempt to decode an invalid string. If strict is False then literal
    control characters are allowed in the string.

    Returns a tuple of the decoded string and the index of the character in s
    after the end quote."""
    chunks = []
    _append = chunks.append
    begin = end - 1
    while 1:
        chunk = _m(s, end)
        if chunk is None:
            raise JSONDecodeError("Unterminated string starting at", s, begin)
        end = chunk.end()
        content, terminator = chunk.groups()
        # Content is contains zero or more unescaped string characters
        if content:
            _append(content)
        # Terminator is the end of string, a literal control character,
        # or a backslash denoting that an escape sequence follows
        if terminator == '"':
            break
        elif terminator != '\\':
            if strict:
                #msg = "Invalid control character %r at" % (terminator,)
                msg = "Invalid control character {0!r} at".format(terminator)
                raise JSONDecodeError(msg, s, end)
            else:
                _append(terminator)
                continue
        try:
            esc = s[end]
        except IndexError:
            raise JSONDecodeError("Unterminated string starting at",
                                  s, begin) from None
        # If not a unicode escape sequence, must be in the lookup table
        if esc != 'u':
            try:
                char = _b[esc]
            except KeyError:
                msg = "Invalid \\escape: {0!r}".format(esc)
                raise JSONDecodeError(msg, s, end)
            end += 1
        else:
            uni = _decode_uXXXX(s, end)
            end += 5
            if 0xd800 <= uni <= 0xdbff and s[end:end + 2] == '\\u':
                uni2 = _decode_uXXXX(s, end + 1)
                if 0xdc00 <= uni2 <= 0xdfff:
                    uni = 0x10000 + (((uni - 0xd800) << 10) | (uni2 - 0xdc00))
                    end += 6
            char = chr(uni)
        _append(char)
    return ''.join(chunks), end


# Use speedup if available
scanstring = c_scanstring or py_scanstring

WHITESPACE = re.compile(r'[ \t\n\r]*', FLAGS)
WHITESPACE_STR = ' \t\n\r'


def JSONObject(s_and_end, strict, scan_once, object_hook, object_pairs_hook,
               memo=None, _w=WHITESPACE.match, _ws=WHITESPACE_STR):
    s, end = s_and_end
    pairs = []
    pairs_append = pairs.append
    # Backwards compatibility
    if memo is None:
        memo = {}
    memo_get = memo.setdefault
    # Use a slice to prevent IndexError from being raised, the following
    # check will raise a more specific ValueError if the string is empty
    nextchar = s[end:end + 1]
    # Normally we expect nextchar == '"'
    if nextchar != '"':
        if nextchar in _ws:
            end = _w(s, end).end()
            nextchar = s[end:end + 1]
        # Trivial empty object
        if nextchar == '}':
            if object_pairs_hook is not None:
                result = object_pairs_hook(pairs)
                return result, end + 1
            pairs = {}
            if object_hook is not None:
                pairs = object_hook(pairs)
            return pairs, end + 1
        elif nextchar != '"':
            raise JSONDecodeError(
                "Expecting property name enclosed in double quotes", s, end)
    end += 1
    while True:
        key, end = scanstring(s, end, strict)
        key = memo_get(key, key)
        # To skip some function call overhead we optimize the fast paths where
        # the JSON key separator is ": " or just ":".
        if s[end:end + 1] != ':':
            end = _w(s, end).end()
            if s[end:end + 1] != ':':
                raise JSONDecodeError("Expecting ':' delimiter", s, end)
        end += 1

        try:
            if s[end] in _ws:
                end += 1
                if s[end] in _ws:
                    end = _w(s, end + 1).end()
        except IndexError:
            pass

        try:
            value, end = scan_once(s, end)
        except StopIteration as err:
            raise JSONDecodeError("Expecting value", s, err.value) from None
        pairs_append((key, value))
        try:
            nextchar = s[end]
            if nextchar in _ws:
                end = _w(s, end + 1).end()
                nextchar = s[end]
        except IndexError:
            nextchar = ''
        end += 1

        if nextchar == '}':
            break
        elif nextchar != ',':
            raise JSONDecodeError("Expecting ',' delimiter", s, end - 1)
        comma_idx = end - 1
        end = _w(s, end).end()
        nextchar = s[end:end + 1]
        end += 1
        if nextchar != '"':
            if nextchar == '}':
                raise JSONDecodeError("Illegal trailing comma before end of object", s, comma_idx)
            raise JSONDecodeError(
                "Expecting property name enclosed in double quotes", s, end - 1)
    if object_pairs_hook is not None:
        result = object_pairs_hook(pairs)
        return result, end
    pairs = dict(pairs)
    if object_hook is not None:
        pairs = object_hook(pairs)
    return pairs, end

def JSONArray(s_and_end, scan_once, _w=WHITESPACE.match, _ws=WHITESPACE_STR):
    s, end = s_and_end
    values = []
    nextchar = s[end:end + 1]
    if nextchar in _ws:
        end = _w(s, end + 1).end()
        nextchar = s[end:end + 1]
    # Look-ahead for trivial empty array
    if nextchar == ']':
        return values, end + 1
    _append = values.append
    while True:
        try:
            value, end = scan_once(s, end)
        except StopIteration as err:
            raise JSONDecodeError("Expecting value", s, err.value) from None
        _append(value)
        nextchar = s[end:end + 1]
        if nextchar in _ws:
            end = _w(s, end + 1).end()
            nextchar = s[end:end + 1]
        end += 1
        if nextchar == ']':
            break
        elif nextchar != ',':
            raise JSONDecodeError("Expecting ',' delimiter", s, end - 1)
        comma_idx = end - 1
        try:
            if s[end] in _ws:
                end += 1
                if s[end] in _ws:
                    end = _w(s, end + 1).end()
            nextchar = s[end:end + 1]
        except IndexError:
            pass
        if nextchar == ']':
            raise JSONDecodeError("Illegal trailing comma before end of array", s, comma_idx)

    return values, end


class JSONDecoder(object):
    """Simple JSON <https://json.org> decoder

    Performs the following translations in decoding by default:

    +---------------+-------------------+
    | JSON          | Python            |
    +===============+===================+
    | object        | dict              |
    +---------------+-------------------+
    | array         | list              |
    +---------------+-------------------+
    | string        | str               |
    +---------------+-------------------+
    | number (int)  | int               |
    +---------------+-------------------+
    | number (real) | float             |
    +---------------+-------------------+
    | true          | True              |
    +---------------+-------------------+
    | false         | False             |
    +---------------+-------------------+
    | null          | None              |
    +---------------+-------------------+

    It also understands ``NaN``, ``Infinity``, and ``-Infinity`` as
    their corresponding ``float`` values, which is outside the JSON spec.

    """

    def __init__(self, *, object_hook=None, parse_float=None,
            parse_int=None, parse_constant=None, strict=True,
            object_pairs_hook=None):
        """``object_hook``, if specified, will be called with the result
        of every JSON object decoded and its return value will be used in
        place of the given ``dict``.  This can be used to provide custom
        deserializations (e.g. to support JSON-RPC class hinting).

        ``object_pairs_hook``, if specified will be called with the result
        of every JSON object decoded with an ordered list of pairs.  The
        return value of ``object_pairs_hook`` will be used instead of the
        ``dict``.  This feature can be used to implement custom decoders.
        If ``object_hook`` is also defined, the ``object_pairs_hook`` takes
        priority.

        ``parse_float``, if specified, will be called with the string
        of every JSON float to be decoded. By default this is equivalent to
        float(num_str). This can be used to use another datatype or parser
        for JSON floats (e.g. decimal.Decimal).

        ``parse_int``, if specified, will be called with the string
        of every JSON int to be decoded. By default this is equivalent to
        int(num_str). This can be used to use another datatype or parser
        for JSON integers (e.g. float).

        ``parse_constant``, if specified, will be called with one of the
        following strings: -Infinity, Infinity, NaN.
        This can be used to raise an exception if invalid JSON numbers
        are encountered.

        If ``strict`` is false (true is the default), then control
        characters will be allowed inside strings.  Control characters in
        this context are those with character codes in the 0-31 range,
        including ``'\\t'`` (tab), ``'\\n'``, ``'\\r'`` and ``'\\0'``.
        """
        self.object_hook = object_hook
        self.parse_float = parse_float or float
        self.parse_int = parse_int or int
        self.parse_constant = parse_constant or _CONSTANTS.__getitem__
        self.strict = strict
        self.object_pairs_hook = object_pairs_hook
        self.parse_object = JSONObject
        self.parse_array = JSONArray
        self.parse_string = scanstring
        self.memo = {}
        self.scan_once = scanner.make_scanner(self)


    def decode(self, s, _w=WHITESPACE.match):
        """Return the Python representation of ``s`` (a ``str`` instance
        containing a JSON document).

        """
        obj, end = self.raw_decode(s, idx=_w(s, 0).end())
        end = _w(s, end).end()
        if end != len(s):
            raise JSONDecodeError("Extra data", s, end)
        return obj

    def raw_decode(self, s, idx=0):
        """Decode a JSON document from ``s`` (a ``str`` beginning with
        a JSON document) and return a 2-tuple of the Python
        representation and the index in ``s`` where the document ended.

        This can be used to decode a JSON document from a string that may
        have extraneous data at the end.

        """
        try:
            obj, end = self.scan_once(s, idx)
        except StopIteration as err:
            raise JSONDecodeError("Expecting value", s, err.value) from None
        return obj, end



# Copyright 2010 Gentoo Foundation
# Distributed under the terms of the GNU General Public License v2

from portage.tests import TestCase
from portage.tests.resolver.ResolverPlayground import (
    ResolverPlayground,
    ResolverPlaygroundTestCase,
)


class UseDepDefaultsTestCase(TestCase):
    def testUseDepDefaultse(self):
        ebuilds = {
            "dev-libs/A-1": {
                "DEPEND": "dev-libs/B[foo]",
                "RDEPEND": "dev-libs/B[foo]",
                "EAPI": "2",
            },
            "dev-libs/A-2": {
                "DEPEND": "dev-libs/B[foo(+)]",
                "RDEPEND": "dev-libs/B[foo(+)]",
                "EAPI": "4",
            },
            "dev-libs/A-3": {
                "DEPEND": "dev-libs/B[foo(-)]",
                "RDEPEND": "dev-libs/B[foo(-)]",
                "EAPI": "4",
            },
            "dev-libs/B-1": {"IUSE": "+foo", "EAPI": "1"},
            "dev-libs/B-2": {},
        }

        test_cases = (
            ResolverPlaygroundTestCase(
                ["=dev-libs/A-1"],
                success=True,
                mergelist=["dev-libs/B-1", "dev-libs/A-1"],
            ),
            ResolverPlaygroundTestCase(
                ["=dev-libs/A-2"],
                success=True,
                mergelist=["dev-libs/B-2", "dev-libs/A-2"],
            ),
            ResolverPlaygroundTestCase(
                ["=dev-libs/A-3"],
                success=True,
                mergelist=["dev-libs/B-1", "dev-libs/A-3"],
            ),
        )

        playground = ResolverPlayground(ebuilds=ebuilds)
        try:
            for test_case in test_cases:
                playground.run_TestCase(test_case)
                self.assertEqual(test_case.test_success, True, test_case.fail_msg)
        finally:
            playground.cleanup()



# Copyright 2010-2020 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

__all__ = ("LicenseManager",)

from portage import os
from portage.dep import ExtendedAtomDict, use_reduce
from portage.exception import InvalidDependString
from portage.localization import _
from portage.util import grabdict, grabdict_package, writemsg
from portage.versions import cpv_getkey, _pkg_str

from portage.package.ebuild._config.helper import ordered_by_atom_specificity


class LicenseManager:
    def __init__(self, locations_manager, user_config=True):
        self._accept_license_str = None
        self._accept_license = None
        self._license_groups = {}
        self._plicensedict = ExtendedAtomDict(dict)
        self._undef_lic_groups = set()

        license_group_locations = list(locations_manager.profile_locations)
        if user_config:
            license_group_locations.append(locations_manager.abs_user_config)

        self._read_license_groups(license_group_locations)

        if user_config:
            for profile in locations_manager.profiles_complex:
                if "profile-license" in profile.profile_formats:
                    self._read_user_config(profile.location)
            self._read_user_config(locations_manager.abs_user_config)

    def _read_user_config(self, abs_user_config):
        licdict = grabdict_package(
            os.path.join(abs_user_config, "package.license"),
            recursive=1,
            allow_wildcard=True,
            allow_repo=True,
            verify_eapi=False,
        )
        for k, v in licdict.items():
            self._plicensedict.setdefault(k.cp, {})[k] = self.expandLicenseTokens(v)

    def _read_license_groups(self, locations):
        for loc in locations:
            for k, v in grabdict(os.path.join(loc, "license_groups")).items():
                self._license_groups.setdefault(k, []).extend(v)

        for k, v in self._license_groups.items():
            self._license_groups[k] = frozenset(v)

    def extract_global_changes(self, old=""):
        ret = old
        atom_license_map = self._plicensedict.get("*/*")
        if atom_license_map is not None:
            v = atom_license_map.pop("*/*", None)
            if v is not None:
                ret = " ".join(v)
                if old:
                    ret = old + " " + ret
                if not atom_license_map:
                    # No tokens left in atom_license_map, remove it.
                    del self._plicensedict["*/*"]
        return ret

    def expandLicenseTokens(self, tokens):
        """Take a token from ACCEPT_LICENSE or package.license and expand it
        if it's a group token (indicated by @) or just return it if it's not a
        group.  If a group is negated then negate all group elements."""
        expanded_tokens = []
        for x in tokens:
            expanded_tokens.extend(self._expandLicenseToken(x, None))
        return expanded_tokens

    def _expandLicenseToken(self, token, traversed_groups):
        negate = False
        rValue = []
        if token.startswith("-"):
            negate = True
            license_name = token[1:]
        else:
            license_name = token
        if not license_name.startswith("@"):
            rValue.append(token)
            return rValue
        group_name = license_name[1:]
        if traversed_groups is None:
            traversed_groups = set()
        license_group = self._license_groups.get(group_name)
        if group_name in traversed_groups:
            writemsg(
                _("Circular license group reference" " detected in '%s'\n")
                % group_name,
                noiselevel=-1,
            )
            rValue.append("@" + group_name)
        elif license_group:
            traversed_groups.add(group_name)
            for l in license_group:
                if l.startswith("-"):
                    writemsg(
                        _("Skipping invalid element %s" " in license group '%s'\n")
                        % (l, group_name),
                        noiselevel=-1,
                    )
                else:
                    rValue.extend(self._expandLicenseToken(l, traversed_groups))
        else:
            if self._license_groups and group_name not in self._undef_lic_groups:
                self._undef_lic_groups.add(group_name)
                writemsg(
                    _("Undefined license group '%s'\n") % group_name, noiselevel=-1
                )
            rValue.append("@" + group_name)
        if negate:
            rValue = ["-" + token for token in rValue]
        return rValue

    def _getPkgAcceptLicense(self, cpv, slot, repo):
        """
        Get an ACCEPT_LICENSE list, accounting for package.license.
        """
        accept_license = self._accept_license
        cp = cpv_getkey(cpv)
        cpdict = self._plicensedict.get(cp)
        if cpdict:
            if not hasattr(cpv, "slot"):
                cpv = _pkg_str(cpv, slot=slot, repo=repo)
            plicence_list = ordered_by_atom_specificity(cpdict, cpv)
            if plicence_list:
                accept_license = list(self._accept_license)
                for x in plicence_list:
                    accept_license.extend(x)
        return accept_license

    def get_pruned_accept_license(self, cpv, use, lic, slot, repo):
        """
        Generate a pruned version of ACCEPT_LICENSE, by intersection with
        LICENSE. This is required since otherwise ACCEPT_LICENSE might be
        too big (bigger than ARG_MAX), causing execve() calls to fail with
        E2BIG errors as in bug #262647.
        """
        try:
            licenses = set(use_reduce(lic, uselist=use, flat=True))
        except InvalidDependString:
            licenses = set()
        licenses.discard("||")

        accept_license = self._getPkgAcceptLicense(cpv, slot, repo)

        if accept_license:
            acceptable_licenses = set()
            for x in accept_license:
                if x == "*":
                    acceptable_licenses.update(licenses)
                elif x == "-*":
                    acceptable_licenses.clear()
                elif x[:1] == "-":
                    acceptable_licenses.discard(x[1:])
                elif x in licenses:
                    acceptable_licenses.add(x)

            licenses = acceptable_licenses
        return " ".join(sorted(licenses))

    def getMissingLicenses(self, cpv, use, lic, slot, repo):
        """
        Take a LICENSE string and return a list of any licenses that the user
        may need to accept for the given package.  The returned list will not
        contain any licenses that have already been accepted.  This method
        can throw an InvalidDependString exception.

        @param cpv: The package name (for package.license support)
        @type cpv: String
        @param use: "USE" from the cpv's metadata
        @type use: String
        @param lic: "LICENSE" from the cpv's metadata
        @type lic: String
        @param slot: "SLOT" from the cpv's metadata
        @type slot: String
        @rtype: List
        @return: A list of licenses that have not been accepted.
        """

        licenses = set(use_reduce(lic, matchall=1, flat=True))
        licenses.discard("||")

        acceptable_licenses = set()
        for x in self._getPkgAcceptLicense(cpv, slot, repo):
            if x == "*":
                acceptable_licenses.update(licenses)
            elif x == "-*":
                acceptable_licenses.clear()
            elif x[:1] == "-":
                acceptable_licenses.discard(x[1:])
            else:
                acceptable_licenses.add(x)

        license_str = lic
        if "?" in license_str:
            use = use.split()
        else:
            use = []

        license_struct = use_reduce(license_str, uselist=use, opconvert=True)
        return self._getMaskedLicenses(license_struct, acceptable_licenses)

    def _getMaskedLicenses(self, license_struct, acceptable_licenses):
        if not license_struct:
            return []
        if license_struct[0] == "||":
            ret = []
            for element in license_struct[1:]:
                if isinstance(element, list):
                    if element:
                        tmp = self._getMaskedLicenses(element, acceptable_licenses)
                        if not tmp:
                            return []
                        ret.extend(tmp)
                else:
                    if element in acceptable_licenses:
                        return []
                    ret.append(element)
            # Return all masked licenses, since we don't know which combination
            # (if any) the user will decide to unmask.
            return ret

        ret = []
        for element in license_struct:
            if isinstance(element, list):
                if element:
                    ret.extend(self._getMaskedLicenses(element, acceptable_licenses))
            else:
                if element not in acceptable_licenses:
                    ret.append(element)
        return ret

    def set_accept_license_str(self, accept_license_str):
        if accept_license_str != self._accept_license_str:
            self._accept_license_str = accept_license_str
            self._accept_license = tuple(
                self.expandLicenseTokens(accept_license_str.split())
            )



# Copyright 2023 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

from portage.tests import TestCase
from portage.tests.resolver.ResolverPlayground import (
    ResolverPlayground,
    ResolverPlaygroundTestCase,
)


class SlotConflictBlockedPruneTestCase(TestCase):
    def testSlotConflictBlockedPrune(self):
        """
        Bug 622270
        Downgrading package (as openssl here) due to un-accepting unstable.
        Dependent package (as rustup here) cannot be rebuilt due to missing
        keyword, so dependee downgrade is cancelled, but other dependents
        (such as xwayland here) are rebuilt nevertheless. This should not
        happen and the rebuilds should be pruned.
        """
        ebuilds = {
            "x11-base/xwayland-23.1.1": {
                "EAPI": "5",
                "RDEPEND": "dev-libs/openssl:=",
            },
            "dev-util/rustup-1.25.2": {
                "EAPI": "5",
                "RDEPEND": "dev-libs/openssl:0=",
                "KEYWORDS": "~x86",
            },
            "dev-libs/openssl-1.1.1u": {
                "EAPI": "5",
                "SLOT": "0/1.1",
            },
            "dev-libs/openssl-3.1.1": {
                "EAPI": "5",
                "SLOT": "0/3",
                "KEYWORDS": "~x86",
            },
        }

        installed = {
            "x11-base/xwayland-23.1.1": {
                "EAPI": "5",
                "RDEPEND": "dev-libs/openssl:0/3=",
            },
            "dev-util/rustup-1.25.2": {
                "EAPI": "5",
                "RDEPEND": "dev-libs/openssl:0/3=",
                "KEYWORDS": "~x86",
            },
            "dev-libs/openssl-3.1.1": {
                "EAPI": "5",
                "SLOT": "0/3",
                "KEYWORDS": "~x86",
            },
        }

        world = ["x11-base/xwayland", "dev-util/rustup"]

        test_cases = (
            ResolverPlaygroundTestCase(
                ["@world"],
                options={"--deep": True, "--update": True, "--verbose": True},
                success=True,
                mergelist=["x11-base/xwayland-23.1.1"],
            ),
        )

        playground = ResolverPlayground(
            ebuilds=ebuilds, installed=installed, world=world
        )
        try:
            for test_case in test_cases:
                playground.run_TestCase(test_case)
                self.assertEqual(test_case.test_success, True, test_case.fail_msg)
        finally:
            playground.cleanup()



# Copyright 2010-2025 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

__all__ = ["fetch"]

import errno
import functools
import glob
import itertools
import json
import logging
import random
import re
import shlex
import stat
import sys
import tempfile
import time

from collections import OrderedDict
from urllib.parse import urlparse
from urllib.parse import quote as urlquote

import portage

from portage import (
    os,
    selinux,
    shutil,
    _encodings,
    _movefile,
    _unicode_encode,
)
from portage.checksum import (
    get_valid_checksum_keys,
    perform_md5,
    verify_all,
    _filter_unaccelarated_hashes,
    _hash_filter,
    _apply_hash_filter,
    checksum_str,
)
from portage.const import BASH_BINARY, CUSTOM_MIRRORS_FILE, GLOBAL_CONFIG_PATH
from portage.data import portage_gid, portage_uid, userpriv_groups
from portage.exception import (
    FileNotFound,
    OperationNotPermitted,
    PortageException,
    TryAgain,
)
from portage.localization import _
from portage.locks import lockfile, unlockfile
from portage.output import colorize, EOutput
from portage.util import (
    apply_recursive_permissions,
    apply_secpass_permissions,
    ensure_dirs,
    grabdict,
    varexpand,
    writemsg,
    writemsg_level,
    writemsg_stdout,
)
from portage.util.futures import asyncio
from portage.process import spawn

_download_suffix = ".__download__"

_userpriv_spawn_kwargs = (
    ("uid", portage_uid),
    ("gid", portage_gid),
    ("groups", userpriv_groups),
    ("umask", 0o02),
)


def _hide_url_passwd(url):
    return re.sub(r"//([^:\s]+):[^@\s]+@", r"//\1:*password*@", url)


def _want_userfetch(settings):
    """
    Check if it's desirable to drop privileges for userfetch.

    @param settings: portage config
    @type settings: portage.package.ebuild.config.config
    @return: True if desirable, False otherwise
    """
    return (
        "userfetch" in settings.features
        and portage.data.secpass >= 2
        and os.getuid() == 0
    )


def _drop_privs_userfetch(settings):
    """
    Drop privileges for userfetch, and update portage.data.secpass
    to correspond to the new privilege level.
    """
    spawn_kwargs = dict(_userpriv_spawn_kwargs)
    try:
        asyncio.run(_ensure_distdir(settings, settings["DISTDIR"]))
    except PortageException:
        if not os.path.isdir(settings["DISTDIR"]):
            raise
    os.setgid(int(spawn_kwargs["gid"]))
    os.setgroups(spawn_kwargs["groups"])
    os.setuid(int(spawn_kwargs["uid"]))
    os.umask(spawn_kwargs["umask"])
    portage.data.secpass = 1
    try:
        os.chdir(settings.get("HOME", "/"))
    except Exception:
        os.chdir("/")


def _spawn_fetch(settings, args, **kwargs):
    """
    Spawn a process with appropriate settings for fetching, including
    userfetch and selinux support.
    """

    global _userpriv_spawn_kwargs

    # Redirect all output to stdout since some fetchers like
    # wget pollute stderr (if portage detects a problem then it
    # can send it's own message to stderr).
    if "fd_pipes" not in kwargs:
        kwargs["fd_pipes"] = {
            0: portage._get_stdin().fileno(),
            1: sys.__stdout__.fileno(),
            2: sys.__stdout__.fileno(),
        }

    logname = None
    if (
        "userfetch" in settings.features
        and os.getuid() == 0
        and portage_gid
        and portage_uid
        and hasattr(os, "setgroups")
    ):
        kwargs.update(_userpriv_spawn_kwargs)
        logname = portage.data._portage_username

    spawn_func = spawn

    if settings.selinux_enabled():
        spawn_func = selinux.spawn_wrapper(spawn_func, settings["PORTAGE_FETCH_T"])

        # bash is an allowed entrypoint, while most binaries are not
        if args[0] != BASH_BINARY:
            args = [BASH_BINARY, "-c", 'exec "$@"', args[0]] + args

    # Ensure that EBUILD_PHASE is set to fetch, so that config.environ()
    # does not filter the calling environment (which may contain needed
    # proxy variables, as in bug #315421).
    phase_backup = settings.get("EBUILD_PHASE")
    settings["EBUILD_PHASE"] = "fetch"
    env = settings.environ()
    if logname is not None:
        env["LOGNAME"] = logname
    try:
        rval = spawn_func(args, env=env, **kwargs)
    finally:
        if phase_backup is None:
            settings.pop("EBUILD_PHASE", None)
        else:
            settings["EBUILD_PHASE"] = phase_backup

    return rval


# Instrumentation hooks for use by unit tests.
_async_spawn_fetch_pre_wait = None
_async_spawn_fetch_post_terminate = None


async def _async_spawn_fetch(settings, args, **kwargs):
    kwargs["returnproc"] = True
    proc = _spawn_fetch(settings, args, **kwargs)
    try:
        if _async_spawn_fetch_pre_wait is not None:
            _async_spawn_fetch_pre_wait(proc)
        return await proc.wait()
    except asyncio.CancelledError:
        proc.terminate()
        if _async_spawn_fetch_post_terminate is not None:
            _async_spawn_fetch_post_terminate(proc)
        raise


_async_spawn_fetch.__doc__ = _spawn_fetch.__doc__
_async_spawn_fetch.__doc__ += """
    This function is a coroutine.
"""

_userpriv_test_write_file_cache = {}
_userpriv_test_write_cmd_script = (
    ">> %(file_path)s 2>/dev/null ; rval=$? ; " + "rm -f  %(file_path)s ; exit $rval"
)


async def _userpriv_test_write_file(settings, file_path):
    """
    Drop privileges and try to open a file for writing. The file may or
    may not exist, and the parent directory is assumed to exist. The file
    is removed before returning.

    @param settings: A config instance which is passed to _spawn_fetch()
    @param file_path: A file path to open and write.
    @return: True if write succeeds, False otherwise.
    """

    global _userpriv_test_write_file_cache, _userpriv_test_write_cmd_script
    rval = _userpriv_test_write_file_cache.get(file_path)
    if rval is not None:
        return rval

    # Optimize away the spawn when privileges do not need to be dropped.
    if not _want_userfetch(settings):
        rval = os.access(os.path.dirname(file_path), os.W_OK)
        _userpriv_test_write_file_cache[file_path] = rval
        return rval

    args = [
        BASH_BINARY,
        "-c",
        _userpriv_test_write_cmd_script % {"file_path": shlex.quote(file_path)},
    ]

    returncode = await _async_spawn_fetch(settings, args)

    rval = returncode == os.EX_OK
    _userpriv_test_write_file_cache[file_path] = rval
    return rval


async def _ensure_distdir(settings, distdir):
    """
    Ensure that DISTDIR exists with appropriate permissions.

    @param settings: portage config
    @type settings: portage.package.ebuild.config.config
    @param distdir: DISTDIR path
    @type distdir: str
    @raise PortageException: portage.exception wrapper exception
    """
    from portage.util.install_mask import _raise_exc

    global _userpriv_test_write_file_cache
    dirmode = 0o070
    filemode = 0o60
    modemask = 0o2
    dir_gid = portage_gid
    if "FAKED_MODE" in settings:
        # When inside fakeroot, directories with portage's gid appear
        # to have root's gid. Therefore, use root's gid instead of
        # portage's gid to avoid spurious permissions adjustments
        # when inside fakeroot.
        dir_gid = 0

    userfetch = portage.data.secpass >= 2 and "userfetch" in settings.features
    userpriv = portage.data.secpass >= 2 and "userpriv" in settings.features
    write_test_file = os.path.join(distdir, ".__portage_test_write__")

    if await _userpriv_test_write_file(settings, write_test_file):
        return

    _userpriv_test_write_file_cache.pop(write_test_file, None)

    already_exists = os.path.isdir(distdir)

    if ensure_dirs(distdir, gid=dir_gid, mode=dirmode, mask=modemask):
        if not already_exists:
            # The directory has just been created
            # and therefore it must be empty.
            return
        writemsg(
            _("Adjusting permissions recursively: '%s'\n") % distdir, noiselevel=-1
        )
        if not apply_recursive_permissions(
            distdir,
            gid=dir_gid,
            dirmode=dirmode,
            dirmask=modemask,
            filemode=filemode,
            filemask=modemask,
            onerror=_raise_exc,
        ):
            raise OperationNotPermitted(
                _("Failed to apply recursive permissions for the portage group.")
            )


def _checksum_failure_temp_file(settings, distdir, basename):
    """
    First try to find a duplicate temp file with the same checksum and return
    that filename if available. Otherwise, use mkstemp to create a new unique
    filename._checksum_failure_.$RANDOM, rename the given file, and return the
    new filename. In any case, filename will be renamed or removed before this
    function returns a temp filename.
    """

    filename = os.path.join(distdir, basename)
    if basename.endswith(_download_suffix):
        normal_basename = basename[: -len(_download_suffix)]
    else:
        normal_basename = basename
    size = os.stat(filename).st_size
    checksum = None
    tempfile_re = re.compile(re.escape(normal_basename) + r"\._checksum_failure_\..*")
    for temp_filename in os.listdir(distdir):
        if not tempfile_re.match(temp_filename):
            continue
        temp_filename = os.path.join(distdir, temp_filename)
        try:
            if size != os.stat(temp_filename).st_size:
                continue
        except OSError:
            continue
        try:
            temp_checksum = perform_md5(temp_filename)
        except FileNotFound:
            # Apparently the temp file disappeared. Let it go.
            continue
        if checksum is None:
            checksum = perform_md5(filename)
        if checksum == temp_checksum:
            os.unlink(filename)
            return temp_filename

    fd, temp_filename = tempfile.mkstemp(
        "", normal_basename + "._checksum_failure_.", distdir
    )
    os.close(fd)
    _movefile(filename, temp_filename, mysettings=settings)
    return temp_filename


def _check_digests(filename, digests, show_errors=1):
    """
    Check digests and display a message if an error occurs.
    @return True if all digests match, False otherwise.
    """
    verified_ok, reason = verify_all(filename, digests)
    if not verified_ok:
        if show_errors:
            writemsg(
                _("!!! Previously fetched" " file: '%s'\n") % filename, noiselevel=-1
            )
            writemsg(_("!!! Reason: %s\n") % reason[0], noiselevel=-1)
            writemsg(
                _("!!! Got:      %s\n" "!!! Expected: %s\n") % (reason[1], reason[2]),
                noiselevel=-1,
            )
        return False
    return True


def _check_distfile(filename, digests, eout, show_errors=1, hash_filter=None):
    """
    @return a tuple of (match, stat_obj) where match is True if filename
    matches all given digests (if any) and stat_obj is a stat result, or
    None if the file does not exist.
    """
    if digests is None:
        digests = {}
    size = digests.get("size")
    if size is not None and len(digests) == 1:
        digests = None

    try:
        st = os.stat(filename)
    except OSError:
        return (False, None)
    if size is not None and size != st.st_size:
        return (False, st)
    if not digests:
        if size is not None:
            eout.ebegin(_("%s size ;-)") % os.path.basename(filename))
            eout.eend(0)
        elif st.st_size == 0:
            # Zero-byte distfiles are always invalid.
            return (False, st)
    else:
        digests = _filter_unaccelarated_hashes(digests)
        if hash_filter is not None:
            digests = _apply_hash_filter(digests, hash_filter)
        if _check_digests(filename, digests, show_errors=show_errors):
            eout.ebegin(f"{os.path.basename(filename)} {' '.join(sorted(digests))} ;-)")
            eout.eend(0)
        else:
            return (False, st)
    return (True, st)


_fetch_resume_size_re = re.compile(r"(^[\d]+)([KMGTPEZY]?$)")

_size_suffix_map = {
    "": 0,
    "K": 10,
    "M": 20,
    "G": 30,
    "T": 40,
    "P": 50,
    "E": 60,
    "Z": 70,
    "Y": 80,
}


class DistfileName(str):
    """
    The DistfileName type represents a distfile name and associated
    content digests, used by MirrorLayoutConfig and related layout
    implementations.

    The path of a distfile within a layout must be dependent on
    nothing more than the distfile name and its associated content
    digests. For filename-hash layout, path is dependent on distfile
    name alone, and the get_filenames implementation yields strings
    corresponding to distfile names. For content-hash layout, path is
    dependent on content digest alone, and the get_filenames
    implementation yields DistfileName instances whose names are equal
    to content digest values. The content-hash layout simply lacks
    the filename-hash layout's innate ability to translate a distfile
    path to a distfile name, and instead caries an innate ability
    to translate a distfile path to a content digest.

    In order to prepare for a migration from filename-hash to
    content-hash layout, all consumers of the layout get_filenames
    method need to be updated to work with content digests as a
    substitute for distfile names. For example, emirrordist requires
    the --content-db option when working with a content-hash layout,
    which serves as a means to associate distfile names
    with content digest values yielded by the content-hash get_filenames
    implementation.
    """

    def __new__(cls, s, digests=None):
        return str.__new__(cls, s)

    def __init__(self, s, digests=None):
        super().__init__()
        self.digests = {} if digests is None else digests

    def digests_equal(self, other):
        """
        Test if digests compare equal to those of another instance.
        """
        if not isinstance(other, DistfileName):
            return False
        matches = []
        for algo, digest in self.digests.items():
            other_digest = other.digests.get(algo)
            if other_digest is not None:
                if other_digest == digest:
                    matches.append(algo)
                else:
                    return False
        return bool(matches)


class FlatLayout:
    def get_path(self, filename):
        return filename

    def get_filenames(self, distdir):
        from portage.util.install_mask import _raise_exc

        for dirpath, dirnames, filenames in os.walk(distdir, onerror=_raise_exc):
            for filename in filenames:
                try:
                    yield portage._unicode_decode(filename, errors="strict")
                except UnicodeDecodeError:
                    # Ignore it. Distfiles names must have valid UTF8 encoding.
                    pass
            return

    @staticmethod
    def verify_args(args):
        return len(args) == 1


class FilenameHashLayout:
    def __init__(self, algo, cutoffs):
        self.algo = algo
        self.cutoffs = [int(x) for x in cutoffs.split(":")]

    def get_path(self, filename):
        fnhash = checksum_str(filename.encode("utf8"), self.algo)
        ret = ""
        for c in self.cutoffs:
            assert c % 4 == 0
            c = c // 4
            ret += fnhash[:c] + "/"
            fnhash = fnhash[c:]
        return ret + filename

    def get_filenames(self, distdir):
        pattern = ""
        for c in self.cutoffs:
            assert c % 4 == 0
            c = c // 4
            pattern += c * "[0-9a-f]" + "/"
        pattern += "*"
        for x in glob.iglob(
            portage._unicode_encode(os.path.join(distdir, pattern), errors="strict")
        ):
            try:
                yield portage._unicode_decode(x, errors="strict").rsplit("/", 1)[1]
            except UnicodeDecodeError:
                # Ignore it. Distfiles names must have valid UTF8 encoding.
                pass

    @staticmethod
    def verify_args(args):
        if len(args) != 3:
            return False
        if args[1] not in get_valid_checksum_keys():
            return False
        # argsidate cutoffs
        for c in args[2].split(":"):
            try:
                c = int(c)
            except ValueError:
                break
            else:
                if c % 4 != 0:
                    break
        else:
            return True
        return False


class ContentHashLayout(FilenameHashLayout):
    """
    The content-hash layout is identical to the filename-hash layout,
    except for these three differences:

    1) A content digest is used instead of a filename digest.

    2) The final element of the path returned from the get_path method
    corresponds to the complete content digest. The path is a function
    of the content digest alone.

    3) Because the path is a function of content digest alone, the
    get_filenames implementation cannot derive distfiles names from
    paths, so it instead yields DistfileName instances whose names are
    equal to content digest values. The DistfileName documentation
    discusses resulting implications.

    Motivations to use the content-hash layout instead of the
    filename-hash layout may include:

    1) Since the file path is independent of the file name, file
    name collisions cannot occur. This makes the content-hash
    layout suitable for storage of multiple types of files (not
    only gentoo distfiles). For example, it can be used to store
    distfiles for multiple linux distros within the same tree,
    with automatic deduplication based on content digest. This
    layout can be used to store and distribute practically anything
    (including binary packages for example).

    2) Allows multiple revisions for the same distfiles name. An
    existing distfile can be updated, and if a user still has an
    older copy of an ebuild repository (or an overlay), then a user
    can successfully fetch a desired revision of the distfile as
    long as it has not been purged from the mirror.

    3) File integrity data is integrated into the layout itself,
    making it very simple to verify the integrity of any file that
    it contains. The only tool required is an implementation of
    the chosen hash algorithm.
    """

    def get_path(self, filename):
        """
        For content-hash, the path is a function of the content digest alone.
        The final element of the path returned from the get_path method
        corresponds to the complete content digest.
        """
        fnhash = remaining = filename.digests[self.algo]
        ret = ""
        for c in self.cutoffs:
            assert c % 4 == 0
            c = c // 4
            ret += remaining[:c] + "/"
            remaining = remaining[c:]
        return ret + fnhash

    def get_filenames(self, distdir):
        """
        Yields DistfileName instances each with filename corresponding
        to a digest value for self.algo, and which can be compared to
        other DistfileName instances with their digests_equal method.
        """
        for filename in super().get_filenames(distdir):
            yield DistfileName(filename, digests=dict([(self.algo, filename)]))

    @staticmethod
    def verify_args(args, filename=None):
        """
        If the filename argument is given, then supported hash
        algorithms are constrained by digests available in the filename
        digests attribute.

        @param args: layout.conf entry args
        @param filename: filename with digests attribute
        @return: True if args are valid for available digest algorithms,
                        and False otherwise
        """
        if len(args) != 3:
            return False
        if filename is None:
            supported_algos = get_valid_checksum_keys()
        else:
            supported_algos = filename.digests
        algo = args[1].upper()
        if algo not in supported_algos:
            return False
        return FilenameHashLayout.verify_args(args)


class MirrorLayoutConfig:
    """
    Class to read layout.conf from a mirror.
    """

    def __init__(self):
        self.structure = ()

    def read_from_file(self, f):
        from portage.util.configparser import (
            SafeConfigParser,
            read_configs,
            ConfigParserError,
        )

        cp = SafeConfigParser()
        read_configs(cp, [f])
        vals = []
        for i in itertools.count():
            try:
                vals.append(tuple(cp.get("structure", "%d" % i).split()))
            except ConfigParserError:
                break
        self.structure = tuple(vals)

    def serialize(self):
        return self.structure

    def deserialize(self, data):
        self.structure = data

    @staticmethod
    def validate_structure(val, filename=None):
        """
        If the filename argument is given, then supported hash
        algorithms are constrained by digests available in the filename
        digests attribute.

        @param val: layout.conf entry args
        @param filename: filename with digests attribute
        @return: True if args are valid for available digest algorithms,
                and False otherwise
        """
        if val[0] == "flat":
            return FlatLayout.verify_args(val)
        elif val[0] == "filename-hash":
            return FilenameHashLayout.verify_args(val)
        elif val[0] == "content-hash":
            return ContentHashLayout.verify_args(val, filename=filename)
        return False

    def get_best_supported_layout(self, filename=None):
        """
        If the filename argument is given, then acceptable hash
        algorithms are constrained by digests available in the filename
        digests attribute.

        @param filename: filename with digests attribute
        """
        for val in self.structure:
            if self.validate_structure(val, filename=filename):
                if val[0] == "flat":
                    return FlatLayout(*val[1:])
                elif val[0] == "filename-hash":
                    return FilenameHashLayout(*val[1:])
                elif val[0] == "content-hash":
                    return ContentHashLayout(*val[1:])
        # fallback
        return FlatLayout()

    def get_all_layouts(self):
        ret = []
        for val in self.structure:
            if not self.validate_structure(val):
                raise ValueError(f"Unsupported structure: {val}")
            if val[0] == "flat":
                ret.append(FlatLayout(*val[1:]))
            elif val[0] == "filename-hash":
                ret.append(FilenameHashLayout(*val[1:]))
            elif val[0] == "content-hash":
                ret.append(ContentHashLayout(*val[1:]))
        if not ret:
            ret.append(FlatLayout())
        return ret


def get_mirror_url(mirror_url, filename, mysettings, cache_path=None):
    """
    Get correct fetch URL for a given file, accounting for mirror
    layout configuration.

    @param mirror_url: Base URL to the mirror (without '/distfiles')
    @param filename: Filename to fetch
    @param cache_path: Path for mirror metadata cache
    @return: Full URL to fetch
    """
    return asyncio.run(
        async_mirror_url(mirror_url, filename, mysettings, cache_path=cache_path)
    )


async def async_mirror_url(mirror_url, filename, mysettings, cache_path=None):
    from portage.util import atomic_ofstream
    from portage.util.configparser import ConfigParserError

    mirror_conf = MirrorLayoutConfig()

    cache = {}
    if cache_path is not None:
        try:
            with open(cache_path) as f:
                cache = json.load(f)
        except (OSError, ValueError):
            pass

    ts, data = cache.get(mirror_url, (0, None))
    # refresh at least daily
    if ts >= time.time() - 86400:
        mirror_conf.deserialize(data)
    else:
        tmpfile = f".layout.conf.{urlparse(mirror_url).hostname}"
        try:
            if mirror_url[:1] == "/":
                tmpfile = os.path.join(mirror_url, "layout.conf")
                mirror_conf.read_from_file(tmpfile)
            elif await async_fetch(
                {tmpfile: (mirror_url + "/distfiles/layout.conf",)},
                mysettings,
                force=1,
                try_mirrors=0,
            ):
                tmpfile = os.path.join(mysettings["DISTDIR"], tmpfile)
                mirror_conf.read_from_file(tmpfile)
            else:
                raise OSError()
        except (ConfigParserError, OSError, UnicodeDecodeError):
            pass
        else:
            cache[mirror_url] = (time.time(), mirror_conf.serialize())
            if cache_path is not None:
                f = atomic_ofstream(cache_path, "w")
                json.dump(cache, f)
                f.close()

    # For some protocols, urlquote is required for correct behavior,
    # and it must not be used for other protocols like rsync and sftp.
    path = mirror_conf.get_best_supported_layout(filename=filename).get_path(filename)
    if urlparse(mirror_url).scheme in ("ftp", "http", "https"):
        path = urlquote(path)
    if mirror_url[:1] == "/":
        return os.path.join(mirror_url, path)
    else:
        return mirror_url + "/distfiles/" + path


async_mirror_url.__doc__ = get_mirror_url.__doc__
async_mirror_url.__doc__ += """
    This function is a coroutine.
"""


def fetch(
    myuris,
    mysettings,
    listonly=0,
    fetchonly=0,
    locks_in_subdir=".locks",
    use_locks=1,
    try_mirrors=1,
    digests=None,
    allow_missing_digests=True,
    force=False,
):
    """
    Fetch files to DISTDIR and also verify digests if they are available.

    @param myuris: Maps each file name to a tuple of available fetch URIs.
    @type myuris: dict
    @param mysettings: Portage config instance.
    @type mysettings: portage.config
    @param listonly: Only print URIs and do not actually fetch them.
    @type listonly: bool
    @param fetchonly: Do not block for files that are locked by a
            concurrent fetcher process. This means that the function can
            return successfully *before* all files have been successfully
            fetched!
    @type fetchonly: bool
    @param use_locks: Enable locks. This parameter is ineffective if
            FEATURES=distlocks is disabled in the portage config!
    @type use_locks: bool
    @param digests: Maps each file name to a dict of digest types and values.
    @type digests: dict
    @param allow_missing_digests: Enable fetch even if there are no digests
            available for verification.
    @type allow_missing_digests: bool
    @param force: Force download, even when a file already exists in
            DISTDIR. This is most useful when there are no digests available,
            since otherwise download will be automatically forced if the
            existing file does not match the available digests. Also, this
            avoids the need to remove the existing file in advance, which
            makes it possible to atomically replace the file and avoid
            interference with concurrent processes.
    @type force: bool
    @rtype: int
    @return: 1 if successful, 0 otherwise.
    """
    return asyncio.run(
        async_fetch(
            myuris,
            mysettings,
            listonly=listonly,
            fetchonly=fetchonly,
            locks_in_subdir=locks_in_subdir,
            use_locks=use_locks,
            try_mirrors=try_mirrors,
            digests=digests,
            allow_missing_digests=allow_missing_digests,
            force=force,
        )
    )


async def async_fetch(
    myuris,
    mysettings,
    listonly=0,
    fetchonly=0,
    locks_in_subdir=".locks",
    use_locks=1,
    try_mirrors=1,
    digests=None,
    allow_missing_digests=True,
    force=False,
):
    from portage.package.ebuild.config import check_config_instance

    if force and digests:
        # Since the force parameter can trigger unnecessary fetch when the
        # digests match, do not allow force=True when digests are provided.
        raise PortageException(
            _("fetch: force=True is not allowed when digests are provided")
        )

    if not myuris:
        return 1

    features = mysettings.features
    restrict = mysettings.get("PORTAGE_RESTRICT", "").split()
    userfetch = portage.data.secpass >= 2 and "userfetch" in features

    # 'nomirror' is bad/negative logic. You Restrict mirroring, not no-mirroring.
    restrict_mirror = "mirror" in restrict or "nomirror" in restrict
    if restrict_mirror:
        if ("mirror" in features) and ("lmirror" not in features):
            # lmirror should allow you to bypass mirror restrictions.
            # XXX: This is not a good thing, and is temporary at best.
            print(
                _(
                    '>>> "mirror" mode desired and "mirror" restriction found; skipping fetch.'
                )
            )
            return 1

    # Generally, downloading the same file repeatedly from
    # every single available mirror is a waste of bandwidth
    # and time, so there needs to be a cap.
    checksum_failure_max_tries = 5
    v = checksum_failure_max_tries
    try:
        v = int(
            mysettings.get(
                "PORTAGE_FETCH_CHECKSUM_TRY_MIRRORS", checksum_failure_max_tries
            )
        )
    except (ValueError, OverflowError):
        writemsg(
            _(
                "!!! Variable PORTAGE_FETCH_CHECKSUM_TRY_MIRRORS"
                " contains non-integer value: '%s'\n"
            )
            % mysettings["PORTAGE_FETCH_CHECKSUM_TRY_MIRRORS"],
            noiselevel=-1,
        )
        writemsg(
            _("!!! Using PORTAGE_FETCH_CHECKSUM_TRY_MIRRORS " "default value: %s\n")
            % checksum_failure_max_tries,
            noiselevel=-1,
        )
        v = checksum_failure_max_tries
    if v < 1:
        writemsg(
            _(
                "!!! Variable PORTAGE_FETCH_CHECKSUM_TRY_MIRRORS"
                " contains value less than 1: '%s'\n"
            )
            % v,
            noiselevel=-1,
        )
        writemsg(
            _("!!! Using PORTAGE_FETCH_CHECKSUM_TRY_MIRRORS " "default value: %s\n")
            % checksum_failure_max_tries,
            noiselevel=-1,
        )
        v = checksum_failure_max_tries
    checksum_failure_max_tries = v
    del v

    fetch_resume_size_default = "350K"
    fetch_resume_size = mysettings.get("PORTAGE_FETCH_RESUME_MIN_SIZE")
    if fetch_resume_size is not None:
        fetch_resume_size = "".join(fetch_resume_size.split())
        if not fetch_resume_size:
            # If it's undefined or empty, silently use the default.
            fetch_resume_size = fetch_resume_size_default
        match = _fetch_resume_size_re.match(fetch_resume_size)
        if match is None or (match.group(2).upper() not in _size_suffix_map):
            writemsg(
                _(
                    "!!! Variable PORTAGE_FETCH_RESUME_MIN_SIZE"
                    " contains an unrecognized format: '%s'\n"
                )
                % mysettings["PORTAGE_FETCH_RESUME_MIN_SIZE"],
                noiselevel=-1,
            )
            writemsg(
                _("!!! Using PORTAGE_FETCH_RESUME_MIN_SIZE " "default value: %s\n")
                % fetch_resume_size_default,
                noiselevel=-1,
            )
            fetch_resume_size = None
    if fetch_resume_size is None:
        fetch_resume_size = fetch_resume_size_default
        match = _fetch_resume_size_re.match(fetch_resume_size)
    fetch_resume_size = (
        int(match.group(1)) * 2 ** _size_suffix_map[match.group(2).upper()]
    )

    # Behave like the package has RESTRICT="primaryuri" after a
    # couple of checksum failures, to increase the probablility
    # of success before checksum_failure_max_tries is reached.
    checksum_failure_primaryuri = 2
    thirdpartymirrors = mysettings.thirdpartymirrors()

    # In the background parallel-fetch process, it's safe to skip checksum
    # verification of pre-existing files in $DISTDIR that have the correct
    # file size. The parent process will verify their checksums prior to
    # the unpack phase.

    parallel_fetchonly = "PORTAGE_PARALLEL_FETCHONLY" in mysettings
    if parallel_fetchonly:
        fetchonly = 1

    check_config_instance(mysettings)

    custommirrors = grabdict(
        os.path.join(mysettings["PORTAGE_CONFIGROOT"], CUSTOM_MIRRORS_FILE), recursive=1
    )

    if listonly or ("distlocks" not in features):
        use_locks = 0

    distdir_writable = os.access(mysettings["DISTDIR"], os.W_OK)
    fetch_to_ro = 0
    if "skiprocheck" in features:
        fetch_to_ro = 1

    if not distdir_writable and fetch_to_ro:
        if use_locks:
            writemsg(
                colorize(
                    "BAD",
                    _(
                        "!!! For fetching to a read-only filesystem, "
                        "locking should be turned off.\n"
                    ),
                ),
                noiselevel=-1,
            )
            writemsg(
                _(
                    "!!! This can be done by adding -distlocks to "
                    "FEATURES in /etc/portage/make.conf\n"
                ),
                noiselevel=-1,
            )
    # 			use_locks = 0

    local_mirrors = []
    public_mirrors = []
    fsmirrors = []
    if try_mirrors:
        for x in custommirrors.get("local", []):
            if x.startswith("/"):
                fsmirrors.append(x)
            else:
                local_mirrors.append(x)
        for x in mysettings["GENTOO_MIRRORS"].split():
            if not x:
                continue
            if x.startswith("/"):
                fsmirrors.append(x.rstrip("/"))
            else:
                public_mirrors.append(x.rstrip("/"))

    hash_filter = _hash_filter(mysettings.get("PORTAGE_CHECKSUM_FILTER", ""))
    if hash_filter.transparent:
        hash_filter = None
    skip_manifest = mysettings.get("EBUILD_SKIP_MANIFEST") == "1"
    if skip_manifest:
        allow_missing_digests = True
    pkgdir = mysettings.get("O")
    if digests is None and not (pkgdir is None or skip_manifest):
        mydigests = (
            mysettings.repositories.get_repo_for_location(
                os.path.dirname(os.path.dirname(pkgdir))
            )
            .load_manifest(pkgdir, mysettings["DISTDIR"])
            .getTypeDigests("DIST")
        )
    elif digests is None or skip_manifest:
        # no digests because fetch was not called for a specific package
        mydigests = {}
    else:
        mydigests = digests

    ro_distdirs = [
        x
        for x in shlex.split(mysettings.get("PORTAGE_RO_DISTDIRS", ""))
        if os.path.isdir(x)
    ]

    restrict_fetch = "fetch" in restrict
    force_mirror = "force-mirror" in features and not restrict_mirror and try_mirrors

    file_uri_tuples = []
    # Check for 'items' attribute since OrderedDict is not a dict.
    if hasattr(myuris, "items"):
        for myfile, uri_set in myuris.items():
            for myuri in uri_set:
                file_uri_tuples.append(
                    (DistfileName(myfile, digests=mydigests.get(myfile)), myuri)
                )
            if not uri_set:
                file_uri_tuples.append(
                    (DistfileName(myfile, digests=mydigests.get(myfile)), None)
                )
    else:
        for myuri in myuris:
            if urlparse(myuri).scheme:
                file_uri_tuples.append(
                    (
                        DistfileName(
                            os.path.basename(myuri),
                            digests=mydigests.get(os.path.basename(myuri)),
                        ),
                        myuri,
                    )
                )
            else:
                file_uri_tuples.append(
                    (
                        DistfileName(
                            os.path.basename(myuri),
                            digests=mydigests.get(os.path.basename(myuri)),
                        ),
                        None,
                    )
                )

    filedict = OrderedDict()
    primaryuri_dict = {}
    thirdpartymirror_uris = {}
    for myfile, myuri in file_uri_tuples:
        override_mirror = (myuri or "").startswith("mirror+")
        override_fetch = override_mirror or (myuri or "").startswith("fetch+")
        if override_fetch:
            myuri = myuri.partition("+")[2]

        if myfile not in filedict:
            filedict[myfile] = []
            if distdir_writable:
                mirror_cache = os.path.join(mysettings["DISTDIR"], ".mirror-cache.json")
            else:
                mirror_cache = None

            # fetch restriction implies mirror restriction
            # but fetch unrestriction does not grant mirror permission
            file_restrict_mirror = (
                restrict_fetch or restrict_mirror
            ) and not override_mirror

            # With fetch restriction, a normal uri may only be fetched from
            # custom local mirrors (if available).  A mirror:// uri may also
            # be fetched from specific mirrors (effectively overriding fetch
            # restriction, but only for specific mirrors).
            location_lists = [local_mirrors]
            if not file_restrict_mirror:
                location_lists.append(public_mirrors)

            for l in itertools.chain(*location_lists):
                filedict[myfile].append(
                    functools.partial(
                        async_mirror_url, l, myfile, mysettings, mirror_cache
                    )
                )
        if myuri is None:
            continue
        if myuri[:9] == "mirror://":
            eidx = myuri.find("/", 9)
            if eidx != -1:
                mirrorname = myuri[9:eidx]
                path = myuri[eidx + 1 :]

                # Try user-defined mirrors first
                if mirrorname in custommirrors:
                    for cmirr in custommirrors[mirrorname]:
                        filedict[myfile].append(cmirr.rstrip("/") + "/" + path)

                # now try the official mirrors
                if mirrorname in thirdpartymirrors:
                    uris = [
                        locmirr.rstrip("/") + "/" + path
                        for locmirr in thirdpartymirrors[mirrorname]
                    ]
                    random.shuffle(uris)
                    filedict[myfile].extend(uris)
                    thirdpartymirror_uris.setdefault(myfile, []).extend(uris)

                if (
                    mirrorname not in custommirrors
                    and mirrorname not in thirdpartymirrors
                ):
                    writemsg(_("!!! No known mirror by the name: %s\n") % (mirrorname))
            else:
                writemsg(_("Invalid mirror definition in SRC_URI:\n"), noiselevel=-1)
                writemsg(f"  {myuri}\n", noiselevel=-1)
        else:
            if (restrict_fetch and not override_fetch) or force_mirror:
                # Only fetch from specific mirrors is allowed.
                continue
            primaryuris = primaryuri_dict.get(myfile)
            if primaryuris is None:
                primaryuris = []
                primaryuri_dict[myfile] = primaryuris
            primaryuris.append(myuri)

    # Order primaryuri_dict values to match that in SRC_URI.
    for uris in primaryuri_dict.values():
        uris.reverse()

    # Prefer thirdpartymirrors over normal mirrors in cases when
    # the file does not yet exist on the normal mirrors.
    for myfile, uris in thirdpartymirror_uris.items():
        primaryuri_dict.setdefault(myfile, []).extend(uris)

    # Now merge primaryuri values into filedict (includes mirrors
    # explicitly referenced in SRC_URI).
    if "primaryuri" in restrict:
        for myfile, uris in filedict.items():
            filedict[myfile] = primaryuri_dict.get(myfile, []) + uris
    else:
        for myfile in filedict:
            filedict[myfile] += primaryuri_dict.get(myfile, [])

    can_fetch = True

    if listonly:
        can_fetch = False

    if can_fetch and not fetch_to_ro:
        try:
            await _ensure_distdir(mysettings, mysettings["DISTDIR"])
        except PortageException as e:
            if not os.path.isdir(mysettings["DISTDIR"]):
                writemsg(f"!!! {str(e)}\n", noiselevel=-1)
                writemsg(
                    _("!!! Directory Not Found: DISTDIR='%s'\n")
                    % mysettings["DISTDIR"],
                    noiselevel=-1,
                )
                writemsg(_("!!! Fetching will fail!\n"), noiselevel=-1)

    if can_fetch and not fetch_to_ro and not os.access(mysettings["DISTDIR"], os.W_OK):
        writemsg(
            _("!!! No write access to '%s'\n") % mysettings["DISTDIR"], noiselevel=-1
        )
        can_fetch = False

    distdir_writable = can_fetch and not fetch_to_ro
    failed_files = set()
    restrict_fetch_msg = False
    valid_hashes = set(get_valid_checksum_keys())
    valid_hashes.discard("size")

    for myfile in filedict:
        """
        fetched  status
        0        nonexistent
        1        partially downloaded
        2        completely downloaded
        """
        fetched = 0

        orig_digests = mydigests.get(myfile, {})

        if not (allow_missing_digests or listonly):
            verifiable_hash_types = set(orig_digests).intersection(valid_hashes)
            if not verifiable_hash_types:
                expected = " ".join(sorted(valid_hashes))
                got = set(orig_digests)
                got.discard("size")
                got = " ".join(sorted(got))
                reason = (
                    _("Insufficient data for checksum verification"),
                    got,
                    expected,
                )
                writemsg(
                    _("!!! Fetched file: %s VERIFY FAILED!\n") % myfile, noiselevel=-1
                )
                writemsg(_("!!! Reason: %s\n") % reason[0], noiselevel=-1)
                writemsg(
                    _("!!! Got:      %s\n!!! Expected: %s\n") % (reason[1], reason[2]),
                    noiselevel=-1,
                )

                if fetchonly:
                    failed_files.add(myfile)
                    continue
                else:
                    return 0

        size = orig_digests.get("size")
        if size == 0:
            # Zero-byte distfiles are always invalid, so discard their digests.
            del mydigests[myfile]
            orig_digests.clear()
            size = None
        pruned_digests = orig_digests
        if parallel_fetchonly:
            pruned_digests = {}
            if size is not None:
                pruned_digests["size"] = size

        myfile_path = os.path.join(mysettings["DISTDIR"], myfile)
        download_path = myfile_path if fetch_to_ro else myfile_path + _download_suffix
        has_space = True
        has_space_superuser = True
        file_lock = None
        if listonly:
            writemsg_stdout("\n", noiselevel=-1)
        else:
            # check if there is enough space in DISTDIR to completely store myfile
            # overestimate the filesize so we aren't bitten by FS overhead
            vfs_stat = None
            if size is not None and hasattr(os, "statvfs"):
                try:
                    vfs_stat = os.statvfs(mysettings["DISTDIR"])
                except OSError as e:
                    writemsg_level(
                        f"!!! statvfs('{mysettings['DISTDIR']}'): {e}\n",
                        noiselevel=-1,
                        level=logging.ERROR,
                    )
                    del e

            if vfs_stat is not None:
                try:
                    mysize = os.stat(myfile_path).st_size
                except OSError as e:
                    if e.errno not in (errno.ENOENT, errno.ESTALE):
                        raise
                    del e
                    mysize = 0
                if (size - mysize + vfs_stat.f_bsize) >= (
                    vfs_stat.f_bsize * vfs_stat.f_bavail
                ):
                    if (size - mysize + vfs_stat.f_bsize) >= (
                        vfs_stat.f_bsize * vfs_stat.f_bfree
                    ):
                        has_space_superuser = False

                    if not has_space_superuser:
                        has_space = False
                    elif portage.data.secpass < 2:
                        has_space = False
                    elif userfetch:
                        has_space = False

            if distdir_writable and use_locks:
                lock_kwargs = {}
                if fetchonly:
                    lock_kwargs["flags"] = os.O_NONBLOCK

                try:
                    file_lock = lockfile(myfile_path, wantnewlockfile=1, **lock_kwargs)
                except TryAgain:
                    writemsg(
                        _(
                            ">>> File '%s' is already locked by "
                            "another fetcher. Continuing...\n"
                        )
                        % myfile,
                        noiselevel=-1,
                    )
                    continue
        try:
            if not listonly:
                eout = EOutput()
                eout.quiet = mysettings.get("PORTAGE_QUIET") == "1"
                match, mystat = _check_distfile(
                    myfile_path, pruned_digests, eout, hash_filter=hash_filter
                )
                if match and not force:
                    # Skip permission adjustment for symlinks, since we don't
                    # want to modify anything outside of the primary DISTDIR,
                    # and symlinks typically point to PORTAGE_RO_DISTDIRS.
                    if distdir_writable and not os.path.islink(myfile_path):
                        try:
                            apply_secpass_permissions(
                                myfile_path,
                                gid=portage_gid,
                                mode=0o664,
                                mask=0o2,
                                stat_cached=mystat,
                            )
                        except PortageException as e:
                            if not os.access(myfile_path, os.R_OK):
                                writemsg(
                                    _("!!! Failed to adjust permissions:" " %s\n")
                                    % str(e),
                                    noiselevel=-1,
                                )
                            del e
                    continue

                # Remove broken symlinks or symlinks to files which
                # _check_distfile did not match above.
                if distdir_writable and mystat is None or os.path.islink(myfile_path):
                    try:
                        os.unlink(myfile_path)
                    except OSError as e:
                        if e.errno not in (errno.ENOENT, errno.ESTALE):
                            raise
                    mystat = None

                if mystat is not None:
                    if stat.S_ISDIR(mystat.st_mode):
                        writemsg_level(
                            _(
                                "!!! Unable to fetch file since "
                                "a directory is in the way: \n"
                                "!!!   %s\n"
                            )
                            % myfile_path,
                            level=logging.ERROR,
                            noiselevel=-1,
                        )
                        return 0

                    if distdir_writable and not force:
                        # Since _check_distfile did not match above, the file
                        # is either corrupt or its identity has changed since
                        # the last time it was fetched, so rename it.
                        temp_filename = _checksum_failure_temp_file(
                            mysettings, mysettings["DISTDIR"], myfile
                        )
                        writemsg_stdout(
                            _("Refetching... " "File renamed to '%s'\n\n")
                            % temp_filename,
                            noiselevel=-1,
                        )

                # Stat the temporary download file for comparison with
                # fetch_resume_size.
                try:
                    mystat = os.stat(download_path)
                except OSError as e:
                    if e.errno not in (errno.ENOENT, errno.ESTALE):
                        raise
                    mystat = None

                if mystat is not None:
                    if mystat.st_size == 0:
                        if distdir_writable:
                            try:
                                os.unlink(download_path)
                            except OSError:
                                pass
                    elif distdir_writable and size is not None:
                        if mystat.st_size < fetch_resume_size and mystat.st_size < size:
                            # If the file already exists and the size does not
                            # match the existing digests, it may be that the
                            # user is attempting to update the digest. In this
                            # case, the digestgen() function will advise the
                            # user to use `ebuild --force foo.ebuild manifest`
                            # in order to force the old digests to be replaced.
                            # Since the user may want to keep this file, rename
                            # it instead of deleting it.
                            writemsg(
                                _(
                                    ">>> Renaming distfile with size "
                                    "%d (smaller than "
                                    "PORTAGE_FETCH_RESU"
                                    "ME_MIN_SIZE)\n"
                                )
                                % mystat.st_size
                            )
                            temp_filename = _checksum_failure_temp_file(
                                mysettings,
                                mysettings["DISTDIR"],
                                os.path.basename(download_path),
                            )
                            writemsg_stdout(
                                _("Refetching... " "File renamed to '%s'\n\n")
                                % temp_filename,
                                noiselevel=-1,
                            )
                        elif mystat.st_size >= size:
                            temp_filename = _checksum_failure_temp_file(
                                mysettings,
                                mysettings["DISTDIR"],
                                os.path.basename(download_path),
                            )
                            writemsg_stdout(
                                _("Refetching... " "File renamed to '%s'\n\n")
                                % temp_filename,
                                noiselevel=-1,
                            )

                if distdir_writable and ro_distdirs:
                    readonly_file = None
                    for x in ro_distdirs:
                        filename = await async_mirror_url(x, myfile, mysettings)
                        match, mystat = _check_distfile(
                            filename, pruned_digests, eout, hash_filter=hash_filter
                        )
                        if match:
                            readonly_file = filename
                            break
                    if readonly_file is not None:
                        try:
                            os.unlink(myfile_path)
                        except OSError as e:
                            if e.errno not in (errno.ENOENT, errno.ESTALE):
                                raise
                            del e
                        os.symlink(readonly_file, myfile_path)
                        continue

                # this message is shown only after we know that
                # the file is not already fetched
                if not has_space:
                    writemsg(
                        _("!!! Insufficient space to store %s in %s\n")
                        % (myfile, mysettings["DISTDIR"]),
                        noiselevel=-1,
                    )

                    if has_space_superuser:
                        writemsg(
                            _(
                                "!!! Insufficient privileges to use "
                                "remaining space.\n"
                            ),
                            noiselevel=-1,
                        )
                        if userfetch:
                            writemsg(
                                _(
                                    '!!! You may set FEATURES="-userfetch"'
                                    " in /etc/portage/make.conf in order to fetch with\n"
                                    "!!! superuser privileges.\n"
                                ),
                                noiselevel=-1,
                            )

                if fsmirrors and not os.path.exists(myfile_path) and has_space:
                    for mydir in fsmirrors:
                        mirror_file = await async_mirror_url(mydir, myfile, mysettings)
                        try:
                            shutil.copyfile(mirror_file, download_path)
                            writemsg(_("Local mirror has file: %s\n") % myfile)
                            break
                        except OSError as e:
                            if e.errno not in (errno.ENOENT, errno.ESTALE):
                                raise
                            del e

                try:
                    mystat = os.stat(download_path)
                except OSError as e:
                    if e.errno not in (errno.ENOENT, errno.ESTALE):
                        raise
                    del e
                else:
                    # Skip permission adjustment for symlinks, since we don't
                    # want to modify anything outside of the primary DISTDIR,
                    # and symlinks typically point to PORTAGE_RO_DISTDIRS.
                    if not os.path.islink(download_path):
                        try:
                            apply_secpass_permissions(
                                download_path,
                                gid=portage_gid,
                                mode=0o664,
                                mask=0o2,
                                stat_cached=mystat,
                            )
                        except PortageException as e:
                            if not os.access(download_path, os.R_OK):
                                writemsg(
                                    _("!!! Failed to adjust permissions:" " %s\n")
                                    % (e,),
                                    noiselevel=-1,
                                )

                    # If the file is empty then it's obviously invalid. Remove
                    # the empty file and try to download if possible.
                    if mystat.st_size == 0:
                        if distdir_writable:
                            try:
                                os.unlink(download_path)
                            except OSError:
                                pass
                    elif not orig_digests:
                        # We don't have a digest, and the temporary file exists.
                        if not force:
                            # Try to resume this download when full
                            # download has not been explicitly forced.
                            fetched = 1
                    else:
                        if (
                            mydigests[myfile].get("size") is not None
                            and mystat.st_size < mydigests[myfile]["size"]
                            and not restrict_fetch
                        ):
                            fetched = 1  # Try to resume this download.
                        elif (
                            parallel_fetchonly
                            and mystat.st_size == mydigests[myfile]["size"]
                        ):
                            eout = EOutput()
                            eout.quiet = mysettings.get("PORTAGE_QUIET") == "1"
                            eout.ebegin(f"{myfile} size ;-)")
                            eout.eend(0)
                            continue
                        else:
                            digests = _filter_unaccelarated_hashes(mydigests[myfile])
                            if hash_filter is not None:
                                digests = _apply_hash_filter(digests, hash_filter)
                            verified_ok, reason = verify_all(download_path, digests)
                            if not verified_ok:
                                writemsg(
                                    _("!!! Previously fetched" " file: '%s'\n")
                                    % myfile,
                                    noiselevel=-1,
                                )
                                writemsg(
                                    _("!!! Reason: %s\n") % reason[0], noiselevel=-1
                                )
                                writemsg(
                                    _("!!! Got:      %s\n" "!!! Expected: %s\n")
                                    % (reason[1], reason[2]),
                                    noiselevel=-1,
                                )
                                if reason[0] == _(
                                    "Insufficient data for checksum verification"
                                ):
                                    return 0
                                if distdir_writable:
                                    temp_filename = _checksum_failure_temp_file(
                                        mysettings,
                                        mysettings["DISTDIR"],
                                        os.path.basename(download_path),
                                    )
                                    writemsg_stdout(
                                        _("Refetching... " "File renamed to '%s'\n\n")
                                        % temp_filename,
                                        noiselevel=-1,
                                    )
                            else:
                                if not fetch_to_ro:
                                    _movefile(
                                        download_path,
                                        myfile_path,
                                        mysettings=mysettings,
                                    )
                                eout = EOutput()
                                eout.quiet = (
                                    mysettings.get("PORTAGE_QUIET", None) == "1"
                                )
                                if digests:
                                    digests = list(digests)
                                    digests.sort()
                                    eout.ebegin(f"{myfile} {' '.join(digests)} ;-)")
                                    eout.eend(0)
                                continue  # fetch any remaining files

            # Create a reversed list since that is optimal for list.pop().
            uri_list = filedict[myfile][:]
            uri_list.reverse()
            checksum_failure_count = 0
            tried_locations = set()
            while uri_list:
                loc = uri_list.pop()
                if isinstance(loc, functools.partial):
                    loc = await loc()
                # Eliminate duplicates here in case we've switched to
                # "primaryuri" mode on the fly due to a checksum failure.
                if loc in tried_locations:
                    continue
                tried_locations.add(loc)
                if listonly:
                    writemsg_stdout(loc + " ", noiselevel=-1)
                    fetched = 2
                    continue
                # allow different fetchcommands per protocol
                protocol = loc[0 : loc.find("://")]

                global_config_path = GLOBAL_CONFIG_PATH
                if portage.const.EPREFIX:
                    global_config_path = os.path.join(
                        portage.const.EPREFIX, GLOBAL_CONFIG_PATH.lstrip(os.sep)
                    )

                missing_file_param = False
                fetchcommand_var = "FETCHCOMMAND_" + protocol.upper()
                fetchcommand = mysettings.get(fetchcommand_var)
                if fetchcommand is None:
                    fetchcommand_var = "FETCHCOMMAND"
                    fetchcommand = mysettings.get(fetchcommand_var)
                    if fetchcommand is None:
                        writemsg_level(
                            _(
                                "!!! %s is unset. It should "
                                "have been defined in\n!!! %s/make.globals.\n"
                            )
                            % (fetchcommand_var, global_config_path),
                            level=logging.ERROR,
                            noiselevel=-1,
                        )
                        return 0
                if "${FILE}" not in fetchcommand:
                    writemsg_level(
                        _(
                            "!!! %s does not contain the required ${FILE}"
                            " parameter.\n"
                        )
                        % fetchcommand_var,
                        level=logging.ERROR,
                        noiselevel=-1,
                    )
                    missing_file_param = True

                resumecommand_var = "RESUMECOMMAND_" + protocol.upper()
                resumecommand = mysettings.get(resumecommand_var)
                if resumecommand is None:
                    resumecommand_var = "RESUMECOMMAND"
                    resumecommand = mysettings.get(resumecommand_var)
                    if resumecommand is None:
                        writemsg_level(
                            _(
                                "!!! %s is unset. It should "
                                "have been defined in\n!!! %s/make.globals.\n"
                            )
                            % (resumecommand_var, global_config_path),
                            level=logging.ERROR,
                            noiselevel=-1,
                        )
                        return 0
                if "${FILE}" not in resumecommand:
                    writemsg_level(
                        _(
                            "!!! %s does not contain the required ${FILE}"
                            " parameter.\n"
                        )
                        % resumecommand_var,
                        level=logging.ERROR,
                        noiselevel=-1,
                    )
                    missing_file_param = True

                if missing_file_param:
                    writemsg_level(
                        _(
                            "!!! Refer to the make.conf(5) man page for "
                            "information about how to\n!!! correctly specify "
                            "FETCHCOMMAND and RESUMECOMMAND.\n"
                        ),
                        level=logging.ERROR,
                        noiselevel=-1,
                    )
                    if myfile != os.path.basename(loc):
                        return 0

                if not can_fetch:
                    if fetched != 2:
                        try:
                            mysize = os.stat(download_path).st_size
                        except OSError as e:
                            if e.errno not in (errno.ENOENT, errno.ESTALE):
                                raise
                            del e
                            mysize = 0

                        if mysize == 0:
                            writemsg(
                                _("!!! File %s isn't fetched but unable to get it.\n")
                                % myfile,
                                noiselevel=-1,
                            )
                        elif size is None or size > mysize:
                            writemsg(
                                _(
                                    "!!! File %s isn't fully fetched, but unable to complete it\n"
                                )
                                % myfile,
                                noiselevel=-1,
                            )
                        else:
                            writemsg(
                                _(
                                    "!!! File %s is incorrect size, "
                                    "but unable to retry.\n"
                                )
                                % myfile,
                                noiselevel=-1,
                            )
                        return 0
                    continue

                if fetched != 2 and has_space:
                    # we either need to resume or start the download
                    if fetched == 1:
                        try:
                            mystat = os.stat(download_path)
                        except OSError as e:
                            if e.errno not in (errno.ENOENT, errno.ESTALE):
                                raise
                            del e
                            fetched = 0
                        else:
                            if distdir_writable and mystat.st_size < fetch_resume_size:
                                writemsg(
                                    _(
                                        ">>> Deleting distfile with size "
                                        "%d (smaller than "
                                        "PORTAGE_FETCH_RESU"
                                        "ME_MIN_SIZE)\n"
                                    )
                                    % mystat.st_size
                                )
                                try:
                                    os.unlink(download_path)
                                except OSError as e:
                                    if e.errno not in (errno.ENOENT, errno.ESTALE):
                                        raise
                                    del e
                                fetched = 0
                    if fetched == 1:
                        # resume mode:
                        writemsg(_(">>> Resuming download...\n"))
                        locfetch = resumecommand
                        command_var = resumecommand_var
                    else:
                        # normal mode:
                        locfetch = fetchcommand
                        command_var = fetchcommand_var
                    writemsg_stdout(_(">>> Downloading '%s'\n") % _hide_url_passwd(loc))
                    variables = {"URI": loc, "FILE": os.path.basename(download_path)}

                    try:
                        variables["DIGESTS"] = " ".join(
                            [
                                f"{k.lower()}:{v}"
                                for k, v in mydigests[myfile].items()
                                if k != "size"
                            ]
                        )
                    except KeyError:
                        pass

                    for k in ("DISTDIR", "PORTAGE_SSH_OPTS"):
                        v = mysettings.get(k)
                        if v is not None:
                            variables[k] = v

                    myfetch = varexpand(locfetch, mydict=variables)
                    myfetch = shlex.split(myfetch)

                    myret = -1
                    try:
                        myret = await _async_spawn_fetch(mysettings, myfetch)

                    finally:
                        try:
                            apply_secpass_permissions(
                                download_path, gid=portage_gid, mode=0o664, mask=0o2
                            )
                        except FileNotFound:
                            pass
                        except PortageException as e:
                            if not os.access(download_path, os.R_OK):
                                writemsg(
                                    _("!!! Failed to adjust permissions:" " %s\n")
                                    % str(e),
                                    noiselevel=-1,
                                )
                            del e

                    # If the file is empty then it's obviously invalid.  Don't
                    # trust the return value from the fetcher.  Remove the
                    # empty file and try to download again.
                    try:
                        mystat = os.lstat(download_path)
                        if mystat.st_size == 0 or (
                            stat.S_ISLNK(mystat.st_mode)
                            and not os.path.exists(download_path)
                        ):
                            os.unlink(download_path)
                            fetched = 0
                            continue
                    except OSError:
                        pass

                    if mydigests is not None and myfile in mydigests:
                        try:
                            mystat = os.stat(download_path)
                        except OSError as e:
                            if e.errno not in (errno.ENOENT, errno.ESTALE):
                                raise
                            del e
                            fetched = 0
                        else:
                            if stat.S_ISDIR(mystat.st_mode):
                                # This can happen if FETCHCOMMAND erroneously
                                # contains wget's -P option where it should
                                # instead have -O.
                                writemsg_level(
                                    _(
                                        "!!! The command specified in the "
                                        "%s variable appears to have\n!!! "
                                        "created a directory instead of a "
                                        "normal file.\n"
                                    )
                                    % command_var,
                                    level=logging.ERROR,
                                    noiselevel=-1,
                                )
                                writemsg_level(
                                    _(
                                        "!!! Refer to the make.conf(5) "
                                        "man page for information about how "
                                        "to\n!!! correctly specify "
                                        "FETCHCOMMAND and RESUMECOMMAND.\n"
                                    ),
                                    level=logging.ERROR,
                                    noiselevel=-1,
                                )
                                return 0

                            # no exception?  file exists. let digestcheck() report
                            # an appropriately for size or checksum errors

                            # If the fetcher reported success and the file is
                            # too small, it's probably because the digest is
                            # bad (upstream changed the distfile).  In this
                            # case we don't want to attempt to resume. Show a
                            # digest verification failure to that the user gets
                            # a clue about what just happened.
                            if (
                                myret != os.EX_OK
                                and mystat.st_size < mydigests[myfile]["size"]
                            ):
                                # Fetch failed... Try the next one... Kill 404 files though.
                                if (
                                    (mystat[stat.ST_SIZE] < 100000)
                                    and (len(myfile) > 4)
                                    and not (
                                        (myfile[-5:] == ".html")
                                        or (myfile[-4:] == ".htm")
                                    )
                                ):
                                    html404 = re.compile(
                                        "<title>.*(not found|404).*</title>",
                                        re.I | re.M,
                                    )
                                    with open(
                                        _unicode_encode(
                                            download_path,
                                            encoding=_encodings["fs"],
                                            errors="strict",
                                        ),
                                        encoding=_encodings["content"],
                                        errors="replace",
                                    ) as f:
                                        if html404.search(f.read()):
                                            try:
                                                os.unlink(download_path)
                                                writemsg(
                                                    _(
                                                        ">>> Deleting invalid distfile. (Improper 404 redirect from server.)\n"
                                                    )
                                                )
                                                fetched = 0
                                                continue
                                            except OSError:
                                                pass
                                fetched = 1
                                continue
                            if True:
                                # File is the correct size--check the checksums for the fetched
                                # file NOW, for those users who don't have a stable/continuous
                                # net connection. This way we have a chance to try to download
                                # from another mirror...
                                digests = _filter_unaccelarated_hashes(
                                    mydigests[myfile]
                                )
                                if hash_filter is not None:
                                    digests = _apply_hash_filter(digests, hash_filter)
                                verified_ok, reason = verify_all(download_path, digests)
                                if not verified_ok:
                                    writemsg(
                                        _("!!! Fetched file: %s VERIFY FAILED!\n")
                                        % myfile,
                                        noiselevel=-1,
                                    )
                                    writemsg(
                                        _("!!! Reason: %s\n") % reason[0], noiselevel=-1
                                    )
                                    writemsg(
                                        _("!!! Got:      %s\n!!! Expected: %s\n")
                                        % (reason[1], reason[2]),
                                        noiselevel=-1,
                                    )
                                    if reason[0] == _(
                                        "Insufficient data for checksum verification"
                                    ):
                                        return 0
                                    if distdir_writable:
                                        temp_filename = _checksum_failure_temp_file(
                                            mysettings,
                                            mysettings["DISTDIR"],
                                            os.path.basename(download_path),
                                        )
                                        writemsg_stdout(
                                            _(
                                                "Refetching... "
                                                "File renamed to '%s'\n\n"
                                            )
                                            % temp_filename,
                                            noiselevel=-1,
                                        )
                                    fetched = 0
                                    checksum_failure_count += 1
                                    if (
                                        checksum_failure_count
                                        == checksum_failure_primaryuri
                                    ):
                                        # Switch to "primaryuri" mode in order
                                        # to increase the probablility of
                                        # of success.
                                        primaryuris = primaryuri_dict.get(myfile)
                                        if primaryuris:
                                            uri_list.extend(reversed(primaryuris))
                                    if (
                                        checksum_failure_count
                                        >= checksum_failure_max_tries
                                    ):
                                        break
                                else:
                                    if not fetch_to_ro:
                                        _movefile(
                                            download_path,
                                            myfile_path,
                                            mysettings=mysettings,
                                        )
                                    eout = EOutput()
                                    eout.quiet = (
                                        mysettings.get("PORTAGE_QUIET", None) == "1"
                                    )
                                    if digests:
                                        eout.ebegin(
                                            f"{myfile} {' '.join(sorted(digests))} ;-)"
                                        )
                                        eout.eend(0)
                                    fetched = 2
                                    break
                    else:  # no digests available
                        if not myret:
                            if not fetch_to_ro:
                                _movefile(
                                    download_path, myfile_path, mysettings=mysettings
                                )
                            fetched = 2
                            break
                        elif mydigests is not None:
                            writemsg(
                                _("No digest file available and download failed.\n\n"),
                                noiselevel=-1,
                            )
        finally:
            if use_locks and file_lock:
                unlockfile(file_lock)
                file_lock = None

        if listonly:
            writemsg_stdout("\n", noiselevel=-1)
        if fetched != 2:
            if restrict_fetch and not restrict_fetch_msg:
                restrict_fetch_msg = True
                msg = _(
                    "\n!!! %s/%s"
                    " has fetch restriction turned on.\n"
                    "!!! This probably means that this "
                    "ebuild's files must be downloaded\n"
                    "!!! manually.  See the comments in"
                    " the ebuild for more information.\n\n"
                ) % (mysettings["CATEGORY"], mysettings["PF"])
                writemsg_level(msg, level=logging.ERROR, noiselevel=-1)
            elif restrict_fetch:
                pass
            elif listonly:
                pass
            elif not filedict[myfile]:
                writemsg(
                    _("Warning: No mirrors available for file" " '%s'\n") % (myfile),
                    noiselevel=-1,
                )
            else:
                writemsg(
                    _("!!! Couldn't download '%s'. Aborting.\n") % myfile, noiselevel=-1
                )

            if listonly:
                failed_files.add(myfile)
                continue
            elif fetchonly:
                failed_files.add(myfile)
                continue
            return 0
    if failed_files:
        return 0
    return 1


async_fetch.__doc__ = fetch.__doc__
async_fetch.__doc__ += """
    This function is a coroutine.
"""



# Copyright 2012 Gentoo Foundation
# Distributed under the terms of the GNU General Public License v2

from portage.tests import TestCase
from portage.tests.resolver.ResolverPlayground import (
    ResolverPlayground,
    ResolverPlaygroundTestCase,
)


class SlotOperatorUnsatisfiedTestCase(TestCase):
    def testSlotOperatorUnsatisfied(self):
        ebuilds = {
            "app-misc/A-1": {"EAPI": "5", "SLOT": "0/1"},
            "app-misc/A-2": {"EAPI": "5", "SLOT": "0/2"},
            "app-misc/B-0": {
                "EAPI": "5",
                "DEPEND": "app-misc/A:=",
                "RDEPEND": "app-misc/A:=",
            },
        }

        installed = {
            "app-misc/A-2": {"EAPI": "5", "SLOT": "0/2"},
            "app-misc/B-0": {
                "EAPI": "5",
                "DEPEND": "app-misc/A:0/1=",
                "RDEPEND": "app-misc/A:0/1=",
            },
        }

        world = ["app-misc/B"]

        test_cases = (
            # Demonstrate bug #439694, where a broken slot-operator
            # sub-slot dependency needs to trigger a rebuild.
            ResolverPlaygroundTestCase(
                ["@world"],
                options={"--update": True, "--deep": True},
                success=True,
                mergelist=["app-misc/B-0"],
            ),
            # This doesn't trigger a rebuild, since there's no version
            # change to trigger complete graph mode, and initially
            # unsatisfied deps are ignored in complete graph mode anyway.
            ResolverPlaygroundTestCase(
                ["app-misc/A"],
                options={"--oneshot": True},
                success=True,
                mergelist=["app-misc/A-2"],
            ),
        )

        playground = ResolverPlayground(
            ebuilds=ebuilds, installed=installed, world=world, debug=False
        )
        try:
            for test_case in test_cases:
                playground.run_TestCase(test_case)
                self.assertEqual(test_case.test_success, True, test_case.fail_msg)
        finally:
            playground.cleanup()
